/*
 * bltHtTag.c --
 *
 *	This module implements an hierarchy widget for the BLT toolkit.
 *
 * Copyright 1998-1999 Lucent Technologies, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that the copyright notice and warranty
 * disclaimer appear in supporting documentation, and that the names
 * of Lucent Technologies or any of their entities not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 *
 * Lucent Technologies disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and
 * fitness.  In no event shall Lucent Technologies be liable for any
 * special, indirect or consequential damages or any damages
 * whatsoever resulting from loss of use, data or profits, whether in
 * an action of contract, negligence or other tortuous action, arising
 * out of or in connection with the use or performance of this
 * software.
 *
 *	The "hiertable" widget was created by George A. Howlett.
 */

#include "bltInt.h"

#ifndef NO_HIERTABLE

#include "bltHiertable.h"
#include "bltChain.h"
#include "bltSwitch.h"
enum TagTypes { TAG_TYPE_NONE, TAG_TYPE_ALL, TAG_TYPE_TAG };

typedef struct {
    Blt_Uid tagUid;
    Blt_HashEntry *hashPtr;
    Blt_ChainLink *linkPtr;
    Blt_HashTable nodeTable;
} TagInfo;

static Tk_Uid allUid, rootUid;

/*
 *----------------------------------------------------------------------
 *
 * GetTagTable --
 *
 *---------------------------------------------------------------------- 
 */
static Blt_HashTable *
GetTagTable(htabPtr, tagUid)
    Hiertable *htabPtr;
    Blt_Uid tagUid;
{
    Blt_HashEntry *hPtr;

    hPtr = Blt_FindHashEntry(&(htabPtr->tagsTable), tagUid);
    if (hPtr != NULL) {
	TagInfo *tagPtr;
	
	tagPtr = (TagInfo *)Blt_GetHashValue(hPtr);
	return &(tagPtr->nodeTable);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyTagInfo --
 *
 *---------------------------------------------------------------------- 
 */
static void
DestroyTagInfo(htabPtr, tagPtr)
    Hiertable *htabPtr;
    TagInfo *tagPtr;
{
    Blt_DeleteHashEntry(tagPtr->hashPtr);
    Blt_DeleteHashTable(&(tagPtr->nodeTable));
    Blt_ChainDeleteLink(htabPtr->tagChainPtr, tagPtr->linkPtr);
    Blt_FreeUid(tagPtr->tagUid);
    Blt_Free((char *)tagPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * ForgetTag --
 *
 *---------------------------------------------------------------------- 
 */
static int
ForgetTag(htabPtr, string)
    Hiertable *htabPtr;
    char *string;
{
    Blt_HashEntry *hPtr;
    TagInfo *tagPtr;
    Blt_Uid tagUid;

    tagUid = Blt_FindUid(string);
    if (tagUid == NULL) {
	return TCL_ERROR;
    }
    if ((tagUid == allUid) || (tagUid == rootUid)) {
	return TCL_OK;
    }
    hPtr = Blt_FindHashEntry(&(htabPtr->tagsTable), tagUid);
    if (hPtr == NULL) {
	return TCL_ERROR;
    }
    tagPtr = (TagInfo *)Blt_GetHashValue(hPtr);
    DestroyTagInfo(htabPtr, tagPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AddTag --
 *
 *---------------------------------------------------------------------- 
 */
static int
AddTag(htabPtr, node, string)
    Hiertable *htabPtr;
    Blt_TreeNode node;
    char *string;
{
    Blt_HashEntry *hPtr;
    TagInfo *tagPtr;
    int isNew;
    Blt_Uid tagUid;

    tagUid = Blt_FindUid(string);
    if ((tagUid == allUid) || (tagUid == rootUid)) {
	Tcl_AppendResult(htabPtr->interp, "can't add reserved tag \"",
			 string, "\"", (char *)NULL);
	return TCL_ERROR;
    }
    tagUid = Blt_GetUid(string);
    hPtr = Blt_CreateHashEntry(&(htabPtr->tagsTable), tagUid, &isNew);
    assert(hPtr);
    if (isNew) {
	tagPtr = (TagInfo *)Blt_Malloc(sizeof(TagInfo));
	Blt_InitHashTable(&(tagPtr->nodeTable), BLT_ONE_WORD_KEYS);
	Blt_SetHashValue(hPtr, (ClientData)tagPtr);
	tagPtr->hashPtr = hPtr;
	tagPtr->tagUid = Blt_GetUid(string);
	tagPtr->linkPtr = Blt_ChainAppend(htabPtr->tagChainPtr, 
		(ClientData)tagPtr);
    } else {
	tagPtr = (TagInfo *)Blt_GetHashValue(hPtr);
    }
    hPtr = Blt_CreateHashEntry(&(tagPtr->nodeTable), (char *)node, &isNew);
    assert(hPtr);
    if (isNew) {
	Blt_SetHashValue(hPtr, (ClientData)node);
    }
    Blt_FreeUid(tagUid);
    return TCL_OK;
}
    
/*
 *----------------------------------------------------------------------
 *
 * HasTag --
 *
 *---------------------------------------------------------------------- 
 */
static int
HasTag(htabPtr, node, tagUid)
    Hiertable *htabPtr;
    Blt_TreeNode node;
    Blt_Uid tagUid;
{
    Blt_HashEntry *hPtr;
    TagInfo *tagPtr;

    if ((tagUid == rootUid) && (node == Blt_TreeRootNode(htabPtr->tree))) {
	return TRUE;
    }
    if (tagUid == allUid) {
	return TRUE;
    }
    hPtr = Blt_FindHashEntry(&(htabPtr->tagsTable), tagUid);
    if (hPtr == NULL) {
	return FALSE;
    }
    tagPtr = (TagInfo *)Blt_GetHashValue(hPtr);
    hPtr = Blt_FindHashEntry(&(tagPtr->nodeTable), (char *)node);
    if (hPtr == NULL) {
	return FALSE;
    }
    return TRUE;
}

static void
ClearTags(htabPtr, node)
    Hiertable *htabPtr;
    Blt_TreeNode node;
{
    Blt_HashEntry *hPtr;
    TagInfo *tagPtr;
    Blt_ChainLink *linkPtr, *nextPtr;

    for (linkPtr = Blt_ChainFirstLink(htabPtr->tagChainPtr); linkPtr != NULL;
	 linkPtr = nextPtr) {
	nextPtr = Blt_ChainNextLink(linkPtr);

	tagPtr = (TagInfo *)Blt_ChainGetValue(linkPtr);
	hPtr = Blt_FindHashEntry(&(tagPtr->nodeTable), (char *)node);
	if (hPtr != NULL) {
	    Blt_DeleteHashEntry(&(tagPtr->nodeTable), hPtr);
	}
	if (tagPtr->nodeTable.numEntries == 0) {
	    DestroyTagInfo(htabPtr, tagPtr);
	}
    }
}

typedef struct {
    int tagType;
    Blt_TreeNode root;
    Blt_HashSearch cursor;
} TagSearch;

static Blt_TreeNode
FirstTaggedNode(interp, htabPtr, objPtr, cursorPtr)	
    Tcl_Interp *interp;
    Hiertable *htabPtr;
    Tcl_Obj *objPtr;
    TagSearch *cursorPtr;
{
    char *string;
    Blt_TreeNode node;
    Blt_Uid tagUid;
    Blt_TreeNode root;

    root = Blt_TreeRootNode(htabPtr->tree);
    string = Tcl_GetString(objPtr);
    cursorPtr->tagType = TAG_TYPE_NONE;
    cursorPtr->root = root;

    if (isdigit(UCHAR(*string))) {
	if (GetNode(htabPtr, objPtr, &node) != TCL_OK) {
	    return NULL;
	}
	return node;
    } 
    tagUid = Blt_FindUid(string);
    if (tagUid != NULL) {
	if (tagUid == allUid) {
	    cursorPtr->tagType = TAG_TYPE_ALL;
	    return root;
	} else if (tagUid == rootUid)  {
	    return root;
	} else {
	    Blt_HashTable *tablePtr;

	    tablePtr = GetTagTable(htabPtr, tagUid);
	    if (tablePtr != NULL) {
		Blt_HashEntry *hPtr;
		
		hPtr = Blt_FirstHashEntry(tablePtr, &(cursorPtr->cursor)); 
		node = (Blt_TreeNode)Blt_GetHashValue(hPtr);
		cursorPtr->tagType = TAG_TYPE_TAG;
		return node;
	    }
	}
    }
    Tcl_AppendResult(interp, "can't find tag or id \"", string, "\" in ", 
	Blt_TreeName(htabPtr->tree), (char *)NULL);
    return NULL;
}

static Blt_TreeNode
NextTaggedNode(node, cursorPtr)	
    Blt_TreeNode node;
    TagSearch *cursorPtr;
{
    if (cursorPtr->tagType == TAG_TYPE_ALL) {
	return Blt_TreeNextNode(cursorPtr->root, node);
    }
    if (cursorPtr->tagType == TAG_TYPE_TAG) {
	Blt_HashEntry *hPtr;

	hPtr = Blt_NextHashEntry(&(cursorPtr->cursor));
	if (hPtr == NULL) {
	    return NULL;
	}
	return (Blt_TreeNode)Blt_GetHashValue(hPtr);
    }
    return NULL;
}

#define PATTERN_NONE		(0)
#define PATTERN_EXACT		(1<<1)
#define PATTERN_GLOB		(1<<2)
#define PATTERN_REGEXP		(1<<3)
#define PATTERN_MASK		(PATTERN_EXACT | PATTERN_GLOB | PATTERN_REGEXP)
#define MATCH_INVERT		(1<<8)
#define MATCH_LEAFONLY		(1<<4)
#define MATCH_NOCASE		(1<<5)
#define MATCH_PATHNAME		(1<<6)

typedef struct {
    Hiertable *htabPtr;		/* Tree to examine. */
    Tcl_DString *dsPtr;		/* List to accumulate the indices of 
				 * matching nodes. */
    int nMatches;		/* Current number of matches. */

    int flags;			/* See flags definitions above. */

    /* Integer options. */
    int maxMatches;		/* If > 0, stop after this many matches. */
    int maxDepth;		/* If > 0, don't descend more than this
				 * many levels. */
    int order;			/* Order of search: Can be either
				 * TREE_PREORDER, TREE_POSTORDER,
				 * TREE_INORDER, TREE_BREADTHFIRST. */
    /* String options. */
    char *pattern;		/* If non-NULL, pattern to use when 
				 * comparing node names. */
    char *addTag;		/* If non-NULL, tag to add to selected nodes. */

    char **command;		/* Command split into a Tcl list. */

    char *key;			/*  */
    char *tag;			/* Search tag */
    char *tagUid;

} FindData;

static int StringToOrder _ANSI_ARGS_((ClientData clientData,
	Tcl_Interp *interp, char *string, char *record, int offset));

static Blt_SwitchCustom orderSwitch =
{
    StringToOrder, (Blt_SwitchFreeProc *)NULL, (ClientData)0,
};

static Blt_SwitchSpec findSwitches[] = 
{
    {BLT_SWITCH_STRING, "-addtag", Blt_Offset(FindData, addTag), 0},
    {BLT_SWITCH_INT_NONNEGATIVE, "-count", Blt_Offset(FindData, maxMatches), 0},
    {BLT_SWITCH_INT_NONNEGATIVE, "-depth", Blt_Offset(FindData, maxDepth), 0},
    {BLT_SWITCH_STRING, "-exact", Blt_Offset(FindData, pattern), 0},
    {BLT_SWITCH_LIST, "-exec", Blt_Offset(FindData, command), 0},
    {BLT_SWITCH_STRING, "-glob", Blt_Offset(FindData, pattern), 0},
    {BLT_SWITCH_FLAG, "-invert", Blt_Offset(FindData, flags), 0, 0, 
	MATCH_INVERT},
    {BLT_SWITCH_STRING, "-key", Blt_Offset(FindData, key), 0},
    {BLT_SWITCH_FLAG, "-leafonly", Blt_Offset(FindData, flags), 0, 0, 
	MATCH_LEAFONLY},
    {BLT_SWITCH_FLAG, "-nocase", Blt_Offset(FindData, flags), 0, 0, 
	MATCH_NOCASE},
    {BLT_SWITCH_CUSTOM, "-order", Blt_Offset(FindData, order), 0, &orderSwitch},
    {BLT_SWITCH_FLAG, "-path", Blt_Offset(FindData, flags), 0, 0, 
	MATCH_PATHNAME},
    {BLT_SWITCH_STRING, "-regexp", Blt_Offset(FindData, pattern), 0},
    {BLT_SWITCH_STRING, "-tag", Blt_Offset(FindData, tag), 0},
    {BLT_SWITCH_END, NULL, 0, 0}
};

/*ARGSUSED*/
static int
MatchNodeProc(node, clientData, order)
    Blt_TreeNode node;
    ClientData clientData;
    int order;			/* Not used. */
{
    FindData *dataPtr = (FindData *)clientData;
    Tcl_DString dString;
    Hiertable *htabPtr = dataPtr->htabPtr;
    Tcl_Interp *interp = dataPtr->htabPtr->interp;
    int result, invert, pattern;
    char *string;

    if ((dataPtr->flags & MATCH_LEAFONLY) && (!Blt_TreeIsLeaf(node))) {
	return TCL_OK;
    }
    if ((dataPtr->maxDepth >= 0) &&
	(dataPtr->maxDepth < Blt_TreeNodeDepth(htabPtr->tree, node))) {
	return TCL_OK;
    }
    result = TRUE;
    pattern = (dataPtr->flags & PATTERN_MASK);

    Tcl_DStringInit(&dString);
    string = NULL;
    if (dataPtr->key != NULL) {
	Tcl_Obj *objPtr;
	
	if (Blt_TreeGetValue((Tcl_Interp *)NULL, htabPtr->tree, node, 
	     dataPtr->key, &objPtr) == TCL_OK) {
	    string = Tcl_GetString(objPtr);
	} else {
	    result = FALSE;
	}
    } else {	    
	if (dataPtr->flags & MATCH_PATHNAME) {
	    string = Blt_HtGetFullName(htabPtr, NodeToEntry(htabPtr, node), 
		FALSE, &dString);
	} else {
	    string = Blt_TreeNodeLabel(node);
	}
    }
    if (pattern != PATTERN_NONE) {
	if (dataPtr->flags & MATCH_NOCASE) {
	    string = Blt_Strdup(string);
	    strtolower(string);
	}
	switch (pattern) {
	case PATTERN_EXACT:
	    result = (strcmp(string, dataPtr->pattern) == 0);
	    break;
	    
	case PATTERN_GLOB:
	    result = Tcl_StringMatch(string, dataPtr->pattern);
	    break;
	    
	case PATTERN_REGEXP:
	    result = Tcl_RegExpMatch(interp, string, dataPtr->pattern); 
	    break;
	}
	if (dataPtr->flags & MATCH_NOCASE) {
	    Blt_Free(string);
	}
    } 
    Tcl_DStringFree(&dString);
    if ((dataPtr->tagUid != NULL) && 
	(!HasTag(htabPtr, node, dataPtr->tagUid))) {
	result = FALSE;
    }
    invert = (dataPtr->flags & MATCH_INVERT) ? TRUE : FALSE;
    if (result != invert) {
	if (dataPtr->addTag != NULL) {
	    if (AddTag(htabPtr, node, dataPtr->addTag) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
	Tcl_DStringAppendElement(dataPtr->dsPtr, 
				 Blt_Itoa(Blt_TreeNodeId(node)));
	if (dataPtr->command != NULL) {
	    result = Tcl_VarEval(interp, dataPtr->command, " ",
		  Blt_Itoa(Blt_TreeNodeId(node)), (char *)NULL);
	    if (result != TCL_OK) {
		return result;
	    }
	}
	dataPtr->nMatches++;
	if ((dataPtr->maxMatches > 0) && 
	    (dataPtr->nMatches >= dataPtr->maxMatches)) {
	    return TCL_BREAK;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Find2Op --
 *
 *---------------------------------------------------------------------- 
 */
static int
Find2Op(htabPtr, interp, argc, argv)
    Hiertable *htabPtr;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
    FindData data;
    int count, result;
    Tcl_DString dString;
    register int i;
    Entry *entryPtr;

    if (StringToEntry(htabPtr, argv[i], &entryPtr) != TCL_OK) {
	return TCL_ERROR;
    }
    memset(&data, 0, sizeof(data));
    data.maxDepth = -1;
    data.order = TREE_POSTORDER;
    data.flags = 0;

    /* Process switches  */
    if (Blt_ProcessSwitches(interp, findSwitches, argc - 3, argv + 3, 
		     (char *)&data, 0) != TCL_OK) {
	return TCL_ERROR;
    }
    count = 0;
    if (Blt_SwitchChanged(findSwitches, "-glob", (char *)NULL)) {
	count++;
	data.flags &= ~PATTERN_MASK;
	data.flags |= PATTERN_GLOB;
    } 
    if (Blt_SwitchChanged(findSwitches, "-regexp", (char *)NULL)) {
	count++;
	data.flags &= ~PATTERN_MASK;
	data.flags |= PATTERN_REGEXP;
    }
    if (Blt_SwitchChanged(findSwitches, "-exact", (char *)NULL)) {
	count++;
	data.flags &= ~PATTERN_MASK;
	data.flags |= PATTERN_EXACT;
    }
    if ((data.flags & MATCH_NOCASE) && (data.pattern != NULL)) {
	strtolower(data.pattern);
    }
    if (data.maxDepth >= 0) {
	data.maxDepth += Blt_TreeNodeDepth(htabPtr->tree, entryPtr->node);
    }
    if (count > 1) {
	/* Two many patterns supplied. */
    }
    if (data.tag != NULL) {
	data.tagUid = Blt_GetUid(data.tag);
    }
    Tcl_DStringInit(&dString);
    data.dsPtr = &dString;
    data.htabPtr = htabPtr;
    if (data.order == TREE_BREADTHFIRST) {
	result = Blt_TreeApplyBFS(entryPtr->node, MatchNodeProc, 
		(ClientData)&data);
    } else {
	result = Blt_TreeApplyDFS(entryPtr->node, MatchNodeProc, 
		(ClientData)&data, data.order);
    }
    Blt_FreeSwitches(findSwitches, (char *)&data, 0);
    if (data.tagUid != NULL) {
	Blt_FreeUid(data.tagUid);
    }
    if (result == TCL_ERROR) {
	return TCL_ERROR;
    }
    Tcl_DStringResult(interp, &dString);
    return TCL_OK;
}

#endif /* NO_HIERTABLE */

