/*
 * bltImage.c --
 *
 *	This module implements image processing procedures for the BLT
 *	toolkit.
 *
 * Copyright 1997-1998 Lucent Technologies, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that the copyright notice and warranty
 * disclaimer appear in supporting documentation, and that the names
 * of Lucent Technologies any of their entities not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 *
 * Lucent Technologies disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and
 * fitness.  In no event shall Lucent Technologies be liable for any
 * special, indirect or consequential damages or any damages
 * whatsoever resulting from loss of use, data or profits, whether in
 * an action of contract, negligence or other tortuous action, arising
 * out of or in connection with the use or performance of this
 * software.
 */

#include "bltInt.h"
#include "bltImage.h"
#include "bltHash.h"
#include <X11/Xutil.h>
#ifndef WIN32
#include <X11/Xproto.h>
#endif

#define CLAMP(c)	((((c) < 0.0) ? 0.0 : ((c) > 255.0) ? 255.0 : (c)))

#define ROTATE_0	0
#define ROTATE_90	1
#define ROTATE_180	2
#define ROTATE_270	3

/*
 * Each call to Tk_GetImage returns a pointer to one of the following
 * structures, which is used as a token by clients (widgets) that
 * display images.
 */
typedef struct TkImageRec {
    Tk_Window tkwin;		/* Window passed to Tk_GetImage (needed to
				 * "re-get" the image later if the manager
				 * changes). */
    Display *display;		/* Display for tkwin.  Needed because when
				 * the image is eventually freed tkwin may
				 * not exist anymore. */
    struct TkImageMasterRec *masterPtr;
				/* Master for this image (identifiers image
				 * manager, for example). */
    ClientData instanceData;
				/* One word argument to pass to image manager
				 * when dealing with this image instance. */
    Tk_ImageChangedProc *changeProc;
				/* Code in widget to call when image changes
				 * in a way that affects redisplay. */
    ClientData widgetClientData;
				/* Argument to pass to changeProc. */
    struct Image *nextPtr;	/* Next in list of all image instances
				 * associated with the same name. */

} TkImage;

/*
 * For each image master there is one of the following structures,
 * which represents a name in the image table and all of the images
 * instantiated from it.  Entries in mainPtr->imageTable point to
 * these structures.
 */
typedef struct TkImageMasterRec {
    Tk_ImageType *typePtr;	/* Information about image type.  NULL means
				 * that no image manager owns this image:  the
				 * image was deleted. */
    ClientData masterData;	/* One-word argument to pass to image mgr
				 * when dealing with the master, as opposed
				 * to instances. */
    int width, height;		/* Last known dimensions for image. */
    Blt_HashTable *tablePtr;	/* Pointer to hash table containing image
				 * (the imageTable field in some TkMainInfo
				 * structure). */
    Blt_HashEntry *hPtr;	/* Hash entry in mainPtr->imageTable for
				 * this structure (used to delete the hash
				 * entry). */
    TkImage *instancePtr;	/* Pointer to first in list of instances
				 * derived from this name. */
} TkImageMaster;


typedef struct TkPhotoMasterRec TkPhotoMaster;
typedef struct TkColorTableRec TkColorTable;

typedef struct TkPhotoInstanceRec {
    TkPhotoMaster *masterPtr;	/* Pointer to master for image. */
    Display *display;		/* Display for windows using this instance. */
    Colormap colormap;		/* The image may only be used in windows with
				 * this particular colormap. */
    struct TkPhotoInstanceRec *nextPtr;
				/* Pointer to the next instance in the list
				 * of instances associated with this master. */
    int refCount;		/* Number of instances using this structure. */
    Tk_Uid palette;		/* Palette for these particular instances. */
    double gamma;		/* Gamma value for these instances. */
    Tk_Uid defaultPalette;	/* Default palette to use if a palette
				 * is not specified for the master. */
    TkColorTable *colorTablePtr;	/* Pointer to information about colors
				 * allocated for image display in windows
				 * like this one. */
    Pixmap pixels;		/* X pixmap containing dithered image. */
    int width, height;		/* Dimensions of the pixmap. */
    char *error;		/* Error image, used in dithering. */
    XImage *imagePtr;		/* Image structure for converted pixels. */
    XVisualInfo visualInfo;	/* Information about the visual that these
				 * windows are using. */
    GC gc;			/* Graphics context for writing images
				 * to the pixmap. */
} TkPhotoInstance;

/*
 * ----------------------------------------------------------------------
 *
 * Tk_ImageDeleted --
 *
 *	Is there any other way to determine if an image has been
 *	deleted?
 *
 * Results:
 *	Returns 1 if the image has been deleted, 0 otherwise.
 *
 * ----------------------------------------------------------------------
 */
/*LINTLIBRARY*/
int
Tk_ImageIsDeleted(tkImage)
    Tk_Image tkImage;		/* Token for image. */
{
    TkImage *imagePtr = (TkImage *) tkImage;

    if (imagePtr->masterPtr == NULL) {
	return TRUE;
    }
    return (imagePtr->masterPtr->typePtr == NULL);
}

/*LINTLIBRARY*/
Tk_ImageMaster
Tk_ImageGetMaster(tkImage)
    Tk_Image tkImage;		/* Token for image. */
{
    TkImage *imagePtr = (TkImage *)tkImage;

    return (Tk_ImageMaster) imagePtr->masterPtr;
}

/*LINTLIBRARY*/
Tk_ImageType *
Tk_ImageGetType(tkImage)
    Tk_Image tkImage;		/* Token for image. */
{
    TkImage *imagePtr = (TkImage *)tkImage;

    return imagePtr->masterPtr->typePtr;
}

/*LINTLIBRARY*/
Pixmap
Tk_ImageGetPhotoPixmap(tkImage)
    Tk_Image tkImage;		/* Token for image. */
{
    TkImage *imagePtr = (TkImage *)tkImage;

    if (strcmp(imagePtr->masterPtr->typePtr->name, "photo") == 0) {
	TkPhotoInstance *instPtr = (TkPhotoInstance *)imagePtr->instanceData;
	return instPtr->pixels;
    }
    return None;
}

#ifdef WIN32

Pixmap
Blt_PhotoImageMask(tkwin, src)
    Tk_Window tkwin;
    Tk_PhotoImageBlock src;
{
    TkWinBitmap *twdPtr;
    int bytesPerRow;
    int offset, count;
    register int x, y;
    unsigned char *bits;
    unsigned char *srcPtr;

    bytesPerRow = ((src.width + 31) & ~31) / 8;
    bits = (unsigned char *)Blt_Calloc(src.height, bytesPerRow);

#define SetBit(x, y) \
    bits[(bytesPerRow * y) + (x / 8)] |= (0x80 >> (x % 8))

    offset = count = 0;
    for (y = 0; y < src.height; y++) {
	srcPtr = src.pixelPtr + offset;
	for (x = 0; x < src.width; x++) {
	    if (srcPtr[src.offset[3]] == 0x00) {
		SetBit(x, y);
	    } else {
		count++;
	    }
	    srcPtr += src.pixelSize;
	}
	offset += src.pitch;
    }
    if (count > 0) {
	HBITMAP hBitmap;
	BITMAP bm;

	bm.bmType = 0;
	bm.bmWidth = src.width;
	bm.bmHeight = src.height;
	bm.bmWidthBytes = bytesPerRow;
	bm.bmPlanes = 1;
	bm.bmBitsPixel = 1;
	bm.bmBits = bits;
	hBitmap = CreateBitmapIndirect(&bm);

	twdPtr = (TkWinBitmap *)Blt_Malloc(sizeof(TkWinBitmap));
	assert(twdPtr);
	twdPtr->type = TWD_BITMAP;
	twdPtr->handle = hBitmap;
	twdPtr->depth = 1;
	if (Tk_WindowId(tkwin) == None) {
	    twdPtr->colormap = DefaultColormap(Tk_Display(tkwin), 
			 DefaultScreen(Tk_Display(tkwin)));
	} else {
	    twdPtr->colormap = Tk_Colormap(tkwin);
	}
    } else {
	twdPtr = NULL;
    }
    if (bits != NULL) {
	Blt_Free((char *)bits);
    }
    return (Pixmap)twdPtr;
}


Pixmap
Blt_ColorimageMask(tkwin, image)
    Tk_Window tkwin;
    Blt_Colorimage image;
{
    TkWinBitmap *twdPtr;
    int bytesPerRow;
    int count;
    register int x, y;
    unsigned char *bits;
    Pix32 *srcPtr;
    int width, height;

    width = Blt_ColorimageWidth(image);
    height = Blt_ColorimageHeight(image);
    bytesPerRow = ((width + 31) & ~31) / 8;
    bits = (unsigned char *)Blt_Calloc(height, bytesPerRow);

    count = 0;
    srcPtr = Blt_ColorimageData(image);
    for (y = height - 1; y >= 0; y--) {
	for (x = 0; x < width; x++) {
	    if (srcPtr->Alpha == 0x00) {
		SetBit(x, y);
		count++;
	    }
	    srcPtr++;
	}
    }
    if (count > 0) {
	HBITMAP hBitmap;
	BITMAP bm;

	bm.bmType = 0;
	bm.bmWidth = Blt_ColorimageWidth(image);
	bm.bmHeight = Blt_ColorimageHeight(image);
	bm.bmWidthBytes = bytesPerRow;
	bm.bmPlanes = 1;
	bm.bmBitsPixel = 1;
	bm.bmBits = bits;
	hBitmap = CreateBitmapIndirect(&bm);

	twdPtr = (TkWinBitmap *)Blt_Malloc(sizeof(TkWinBitmap));
	assert(twdPtr);
	twdPtr->type = TWD_BITMAP;
	twdPtr->handle = hBitmap;
	twdPtr->depth = 1;
	if (Tk_WindowId(tkwin) == None) {
	    twdPtr->colormap = DefaultColormap(Tk_Display(tkwin), 
			 DefaultScreen(Tk_Display(tkwin)));
	} else {
	    twdPtr->colormap = Tk_Colormap(tkwin);
	}
    } else {
	twdPtr = NULL;
    }
    if (bits != NULL) {
	Blt_Free((char *)bits);
    }
    return (Pixmap)twdPtr;
}

#else

Pixmap
Blt_PhotoImageMask(tkwin, src)
    Tk_Window tkwin;
    Tk_PhotoImageBlock src;
{
    Pixmap bitmap;
    int arraySize, bytes_per_line;
    int offset, count;
    int value, bitMask;
    register int x, y;
    unsigned char *bits;
    unsigned char *srcPtr;
    unsigned char *destPtr;
    unsigned long pixel;

    bytes_per_line = (src.width + 7) / 8;
    arraySize = src.height * bytes_per_line;
    bits = (unsigned char *)Blt_Malloc(sizeof(unsigned char) * arraySize);
    assert(bits);
    destPtr = bits;
    offset = count = 0;
    for (y = 0; y < src.height; y++) {
	value = 0, bitMask = 1;
	srcPtr = src.pixelPtr + offset;
	for (x = 0; x < src.width; /*empty*/ ) {
	    pixel = (srcPtr[src.offset[3]] != 0x00);
	    if (pixel) {
		value |= bitMask;
	    } else {
		count++;	/* Count the number of transparent pixels. */
	    }
	    bitMask <<= 1;
	    x++;
	    if (!(x & 7)) {
		*destPtr++ = (unsigned char)value;
		value = 0, bitMask = 1;
	    }
	    srcPtr += src.pixelSize;
	}
	if (x & 7) {
	    *destPtr++ = (unsigned char)value;
	}
	offset += src.pitch;
    }
    if (count > 0) {
	Tk_MakeWindowExist(tkwin);
	bitmap = XCreateBitmapFromData(Tk_Display(tkwin), Tk_WindowId(tkwin),
	    (char *)bits, (unsigned int)src.width, (unsigned int)src.height);
    } else {
	bitmap = None;		/* Image is opaque. */
    }
    Blt_Free((char *)bits);
    return bitmap;
}

Pixmap
Blt_ColorimageMask(tkwin, image)
    Tk_Window tkwin;
    Blt_Colorimage image;
{
    Pixmap bitmap;
    int arraySize, bytes_per_line;
    int count;
    int value, bitMask;
    register int x, y;
    unsigned char *bits;
    Pix32 *srcPtr;
    unsigned char *destPtr;
    unsigned long pixel;
    int width, height;

    width = Blt_ColorimageWidth(image);
    height = Blt_ColorimageHeight(image);
    bytes_per_line = (width + 7) / 8;
    arraySize = height * bytes_per_line;
    bits = (unsigned char *)Blt_Malloc(sizeof(unsigned char) * arraySize);
    assert(bits);
    destPtr = bits;
    count = 0;
    srcPtr = Blt_ColorimageData(image);
    for (y = 0; y < height; y++) {
	value = 0, bitMask = 1;
	for (x = 0; x < width; /*empty*/ ) {
	    pixel = (srcPtr->Alpha != 0x00);
	    if (pixel) {
		value |= bitMask;
	    } else {
		count++;	/* Count the number of transparent pixels. */
	    }
	    bitMask <<= 1;
	    x++;
	    if (!(x & 7)) {
		*destPtr++ = (unsigned char)value;
		value = 0, bitMask = 1;
	    }
	    srcPtr++;
	}
	if (x & 7) {
	    *destPtr++ = (unsigned char)value;
	}
    }
    if (count > 0) {
	Tk_MakeWindowExist(tkwin);
	bitmap = XCreateBitmapFromData(Tk_Display(tkwin), Tk_WindowId(tkwin),
	    (char *)bits, (unsigned int)width, (unsigned int)height);
    } else {
	bitmap = None;		/* Image is opaque. */
    }
    Blt_Free((char *)bits);
    return bitmap;
}

#endif /*WIN32*/

/*LINTLIBRARY*/
GC
Tk_ImageGetPhotoGC(photoImage)
    Tk_Image photoImage;		/* Token for image. */
{
    TkImage *imagePtr = (TkImage *) photoImage;
    if (strcmp(imagePtr->masterPtr->typePtr->name, "photo") == 0) {
	TkPhotoInstance *instPtr = (TkPhotoInstance *)imagePtr->instanceData;
	return instPtr->gc;
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_CreateColorimage --
 *
 *      Allocates a color image of a designated height and width.
 *
 *	This routine will be augmented with other types of information
 *	such as a color table, etc.
 *
 * Results:
 *      Returns the new color image.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_CreateColorimage(width, height)
    int width, height;		/* Dimensions of new image */
{
    unsigned int nPixels;
    register Pix32 *dataPtr, *endPtr;
    struct Colorimage *colorImgPtr;

    nPixels = width * height;
    dataPtr = (Pix32 *)Blt_Malloc(sizeof(Pix32) * nPixels);
    assert(dataPtr);
    colorImgPtr = (struct Colorimage *)Blt_Malloc(sizeof(struct Colorimage));
    assert(colorImgPtr);
    colorImgPtr->width = width;
    colorImgPtr->height = height;
    colorImgPtr->dataPtr = dataPtr;
    endPtr = dataPtr + nPixels;
    for (/* empty */; dataPtr < endPtr; dataPtr++) {
	dataPtr->value = 0;
	dataPtr->Alpha = (unsigned char)-1;
    }
    return colorImgPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_FreeColorimage --
 *
 *      Deallocates the given color image.
 *
 * Results:
 *      None.
 *
 *----------------------------------------------------------------------
 */
void
Blt_FreeColorimage(colorImgPtr)
    struct Colorimage *colorImgPtr;
{
    Blt_Free((char *)colorImgPtr->dataPtr);
    Blt_Free((char *)colorImgPtr);
}

void
Blt_GammaCorrectColorimage(src, newGamma)
    Blt_Colorimage src;
    double newGamma;
{
    unsigned int nPixels;
    register Pix32 *srcPtr, *endPtr;
    register unsigned int i;
    double value;
    unsigned char table[256];
    double invGamma;

    invGamma = 1.0 / newGamma;
    for (i = 0; i < 256; i++) {
	value = 255.0 * pow((double)i / 255.0, invGamma);
	table[i] = (unsigned char)CLAMP(value);
    }
    nPixels = Blt_ColorimageWidth(src) * Blt_ColorimageHeight(src);
    srcPtr = Blt_ColorimageData(src);
    endPtr = srcPtr + nPixels;
    for (/* empty */; srcPtr < endPtr; srcPtr++) {
	srcPtr->Red = table[srcPtr->Red];
	srcPtr->Green = table[srcPtr->Green];
	srcPtr->Blue = table[srcPtr->Blue];
    }
}

static Region2D *
ColorimageRegion(image, regionPtr)
    Blt_Colorimage image;	/* Image to use as source region. */
    Region2D *regionPtr;	/* Region of source image to copy */
{
    regionPtr->left = regionPtr->top = 0;
    regionPtr->right = Blt_ColorimageWidth(image) - 1;
    regionPtr->bottom = Blt_ColorimageHeight(image) - 1;
    return regionPtr;
}   

Region2D *
Blt_SetRegion(x, y, width, height, regionPtr)
    int x, y, width, height;
    Region2D *regionPtr;	
{
    regionPtr->left = x;
    regionPtr->top = y;
    regionPtr->right = x + width - 1;
    regionPtr->bottom = y + height - 1;
    return regionPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ColorimageToPhoto --
 *
 *      Translates a color image into a Tk photo.
 *
 * Results:
 *      The photo is re-written with the new color image.
 *
 *----------------------------------------------------------------------
 */
void
Blt_ColorimageToPhoto(src, photo)
    Blt_Colorimage src;		/* Image to use as source */
    Tk_PhotoHandle photo;	/* Photo to write color image into */
{
    Tk_PhotoImageBlock dest;
    int width, height;

    width = Blt_ColorimageWidth(src);
    height = Blt_ColorimageHeight(src);

    Tk_PhotoGetImage(photo, &dest);
    dest.pixelSize = sizeof(Pix32);
    dest.pitch = sizeof(Pix32) * width;
    dest.width = width;
    dest.height = height;
    dest.offset[0] = Tk_Offset(Pix32, Red);
    dest.offset[1] = Tk_Offset(Pix32, Green);
    dest.offset[2] = Tk_Offset(Pix32, Blue);
    dest.offset[3] = Tk_Offset(Pix32, Alpha); 
    dest.pixelPtr = (unsigned char *)Blt_ColorimageData(src);
    Tk_PhotoSetSize(photo, width, height);
    Tk_PhotoPutBlock(photo, &dest, 0, 0, width, height);
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_PhotoToColorimage --
 *
 *      Create a photo to a color image.
 *
 * Results:
 *      The new color image is returned.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_PhotoToColorimage(photo, regionPtr)
    Tk_PhotoHandle photo;	/* Source photo image to scale */
    Region2D *regionPtr;	/* Region of source image to copy */
{
    Tk_PhotoImageBlock src;
    Blt_Colorimage image;
    register Pix32 *destPtr;
    register unsigned char *srcData;
    register int offset;
    register int x, y;
    int width, height;
    Region2D region;

    Tk_PhotoGetImage(photo, &src);
    if (regionPtr == NULL) {
	regionPtr = Blt_SetRegion(0, 0, src.width, src.height, &region);
    }
    width = RegionWidth(regionPtr);
    height = RegionHeight(regionPtr);
    image = Blt_CreateColorimage(width, height);
    offset = (regionPtr->left * src.pixelSize) + (regionPtr->top * src.pitch);
    destPtr = Blt_ColorimageData(image);
    if (src.pixelSize == 4) {
        for (y = 0; y < height; y++) {
	    srcData = src.pixelPtr + offset;
	    for (x = 0; x < width; x++) {
	        destPtr->Red = srcData[src.offset[0]];
	        destPtr->Green = srcData[src.offset[1]];
	        destPtr->Blue = srcData[src.offset[2]];
	        destPtr->Alpha = srcData[src.offset[3]];
	        srcData += src.pixelSize;
	        destPtr++;
	    }
	    offset += src.pitch;
        }
    } else if (src.pixelSize == 3) {
        for (y = 0; y < height; y++) {
	    srcData = src.pixelPtr + offset;
	    for (x = 0; x < width; x++) {
	        destPtr->Red = srcData[src.offset[0]];
	        destPtr->Green = srcData[src.offset[1]];
	        destPtr->Blue = srcData[src.offset[2]];
		/* No transparency information */
	        destPtr->Alpha = (unsigned char)-1;
	        srcData += src.pixelSize;
	        destPtr++;
	    }
	    offset += src.pitch;
        }
    } else {
        for (y = 0; y < height; y++) {
	    srcData = src.pixelPtr + offset;
	    for (x = 0; x < width; x++) {
	        destPtr->Red = destPtr->Green = destPtr->Blue = 
		    srcData[src.offset[0]];
		/* No transparency information */
	        destPtr->Alpha = (unsigned char)-1;
	        srcData += src.pixelSize;
	        destPtr++;
	    }
	    offset += src.pitch;
        }
    } 
    return image;
}

#ifdef WIN32

/*
 *----------------------------------------------------------------------
 *
 * Blt_ColorimageToPixmap --
 *
 *      Converts a color image into a pixmap.
 *
 *	Right now this only handles TrueColor visuals.
 *
 * Results:
 *      The new pixmap is returned.
 *
 *----------------------------------------------------------------------
 */
Pixmap
Blt_ColorimageToPixmap(interp, tkwin, image, colorTablePtr)
    Tcl_Interp *interp;
    Tk_Window tkwin;
    Blt_Colorimage image;
    ColorTable *colorTablePtr;	/* Points to array of colormap indices */
{
    HDC pixmapDC;
    TkWinDCState state;
    Display *display;
    int width, height, depth;
    Pixmap pixmap;
    register int x, y;
    register Pix32 *srcPtr;
    COLORREF rgb;

    *colorTablePtr = NULL;
    width = Blt_ColorimageWidth(image);
    height = Blt_ColorimageHeight(image);
    display = Tk_Display(tkwin);
    depth = Tk_Depth(tkwin);

    pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
    pixmapDC = TkWinGetDrawableDC(display, pixmap, &state);

    srcPtr = Blt_ColorimageData(image);
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    rgb = PALETTERGB(srcPtr->Red, srcPtr->Green, srcPtr->Blue);
	    SetPixelV(pixmapDC, x, y, rgb);
	    srcPtr++;
	}
    }
    TkWinReleaseDrawableDC(pixmap, pixmapDC, &state);
    return pixmap;
}

#else

int redAdjust, greenAdjust, blueAdjust;
int redMaskShift, greenMaskShift, blueMaskShift;

/*
 *----------------------------------------------------------------------
 *
 * ShiftCount --
 *
 *	Returns the position of the least significant (low) bit in
 *	the given mask.
 *
 *	For TrueColor and DirectColor visuals, a pixel value is
 *	formed by OR-ing the red, green, and blue colormap indices
 *	into a single 32-bit word.  The visual's color masks tell
 *	you where in the word the indices are supposed to be.  The
 *	masks contain bits only where the index is found.  By counting
 *	the leading zeros in the mask, we know how many bits to shift
 *	to the individual red, green, and blue values to form a pixel.
 *
 * Results:
 *      The number of the least significant bit.
 *
 *----------------------------------------------------------------------
 */
static int
ShiftCount(mask)
    register unsigned int mask;
{
    register int count;

    for (count = 0; count < 32; count++) {
	if (mask & 0x01) {
	    break;
	}
	mask >>= 1;
    }
    return count;
}

/*
 *----------------------------------------------------------------------
 *
 * CountBits --
 *
 *	Returns the number of bits set in the given mask.
 *
 *	Reference: Graphics Gems Volume 2.
 *	
 * Results:
 *      The number of bits to set in the mask.
 *
 *
 *----------------------------------------------------------------------
 */
static int
CountBits(mask)
    register unsigned long mask; /* 32  1-bit tallies */
{
    /* 16  2-bit tallies */
    mask = (mask & 0x55555555) + ((mask >> 1) & (0x55555555));  
    /* 8  4-bit tallies */
    mask = (mask & 0x33333333) + ((mask >> 2) & (0x33333333)); 
    /* 4  8-bit tallies */
    mask = (mask & 0x07070707) + ((mask >> 4) & (0x07070707));  
    /* 2 16-bit tallies */
    mask = (mask & 0x000F000F) + ((mask >> 8) & (0x000F000F));  
    /* 1 32-bit tally */
    mask = (mask & 0x0000001F) + ((mask >> 16) & (0x0000001F));  
    return mask;
}

static void
ComputeMasks(visualPtr)
    Visual *visualPtr;
{
    int count;

    redMaskShift = ShiftCount((unsigned int)visualPtr->red_mask);
    greenMaskShift = ShiftCount((unsigned int)visualPtr->green_mask);
    blueMaskShift = ShiftCount((unsigned int)visualPtr->blue_mask);

    redAdjust = greenAdjust = blueAdjust = 0;
    count = CountBits((unsigned long)visualPtr->red_mask);
    if (count < 8) {
	redAdjust = 8 - count;
    }
    count = CountBits((unsigned long)visualPtr->green_mask);
    if (count < 8) {
	greenAdjust = 8 - count;
    }
    count = CountBits((unsigned long)visualPtr->blue_mask);
    if (count < 8) {
	blueAdjust = 8 - count;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TrueColorPixel --
 *
 *      Computes a pixel index from the 3 component RGB values.
 *
 * Results:
 *      The pixel index is returned.
 *
 *----------------------------------------------------------------------
 */
static INLINE unsigned int
TrueColorPixel(visualPtr, pixelPtr)
    Visual *visualPtr;
    Pix32 *pixelPtr;
{
    unsigned int red, green, blue;

    /*
     * The number of bits per color may be less than eight. For example,
     * 15/16 bit displays (hi-color) use only 5 bits, 8-bit displays
     * use 2 or 3 bits (don't ask me why you'd have an 8-bit TrueColor
     * display). So shift off the least significant bits.
     */
    red = ((unsigned int)pixelPtr->Red >> redAdjust);
    green = ((unsigned int)pixelPtr->Green >> greenAdjust);
    blue = ((unsigned int)pixelPtr->Blue >> blueAdjust);

    /* Shift each color into the proper location of the pixel index. */
    red = (red << redMaskShift) & visualPtr->red_mask;
    green = (green << greenMaskShift) & visualPtr->green_mask;
    blue = (blue << blueMaskShift) & visualPtr->blue_mask;
    return (red | green | blue);
}

/*
 *----------------------------------------------------------------------
 *
 * DirectColorPixel --
 *
 *      Translates the 3 component RGB values into a pixel index.
 *      This differs from TrueColor only in that it first translates
 *	the RGB values through a color table.
 *
 * Results:
 *      The pixel index is returned.
 *
 *----------------------------------------------------------------------
 */
static INLINE unsigned int
DirectColorPixel(colorTabPtr, pixelPtr)
    struct ColorTableRec *colorTabPtr;
    Pix32 *pixelPtr;
{
    unsigned int red, green, blue;

    red = colorTabPtr->red[pixelPtr->Red];
    green = colorTabPtr->green[pixelPtr->Green];
    blue = colorTabPtr->blue[pixelPtr->Blue];
    return (red | green | blue);
}

/*
 *----------------------------------------------------------------------
 *
 * PseudoColorPixel --
 *
 *      Translates the 3 component RGB values into a pixel index.
 *      This differs from TrueColor only in that it first translates
 *	the RGB values through a color table.
 *
 * Results:
 *      The pixel index is returned.
 *
 *----------------------------------------------------------------------
 */
static INLINE unsigned int
PseudoColorPixel(pixelPtr, lut)
    Pix32 *pixelPtr;
    unsigned int *lut;
{
    int red, green, blue;
    int pixel;

    red = (pixelPtr->Red >> 3) + 1;
    green = (pixelPtr->Green >> 3) + 1;
    blue = (pixelPtr->Blue >> 3) + 1;
    pixel = RGBIndex(red, green, blue);
    return lut[pixel];
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ColorimageToPixmap --
 *
 *      Converts a color image into a pixmap.
 *
 *	Right now this only handles TrueColor visuals.
 *
 * Results:
 *      The new pixmap is returned.
 *
 *----------------------------------------------------------------------
 */
Pixmap
Blt_ColorimageToPixmap(interp, tkwin, image, colorTablePtr)
    Tcl_Interp *interp;
    Tk_Window tkwin;
    Blt_Colorimage image;
    ColorTable *colorTablePtr;	/* Points to array of colormap indices */
{
    Display *display;
    int width, height;
    Pixmap pixmap;
    GC pixmapGC;
    Visual *visualPtr;
    XImage *imagePtr;
    int nPixels;

    visualPtr = Tk_Visual(tkwin);
    width = Blt_ColorimageWidth(image);
    height = Blt_ColorimageHeight(image);
    display = Tk_Display(tkwin);

    ComputeMasks(visualPtr);

    *colorTablePtr = NULL;
    imagePtr = XCreateImage(Tk_Display(tkwin), visualPtr, Tk_Depth(tkwin),
	ZPixmap, 0, (char *)NULL, width, height, 32, 0);
    assert(imagePtr);

    nPixels = width * height;
    imagePtr->data = (char *)Blt_Malloc(sizeof(Pix32) * nPixels);
    assert(imagePtr->data);

    imagePtr->byte_order = MSBFirst;	/* Force the byte order */
    imagePtr->bitmap_bit_order = imagePtr->byte_order;
    imagePtr->bytes_per_line = width * sizeof(Pix32);

    switch (visualPtr->class) {
    case TrueColor:
	{
	    register int x, y;
	    register Pix32 *srcPtr;
	    register char *destPtr;
	    unsigned int pixel;
	    int rowOffset;

	    /*
	     * Compute the colormap locations directly from pixel RGB values.
	     */
	    srcPtr = Blt_ColorimageData(image);
	    rowOffset = 0;
	    for (y = 0; y < height; y++) {
		destPtr = imagePtr->data + rowOffset;
		for (x = 0; x < width; x++, srcPtr++) {
		    pixel = TrueColorPixel(visualPtr, srcPtr);
		    switch (imagePtr->bits_per_pixel) {
		    case 32:
			*destPtr++ = (pixel >> 24) & 0xFF;
			/*FALLTHRU*/
		    case 24:
			*destPtr++ = (pixel >> 16) & 0xFF;
			/*FALLTHRU*/
		    case 16:
			*destPtr++ = (pixel >> 8) & 0xFF;
			/*FALLTHRU*/
		    case 8:
			*destPtr++ = pixel & 0xFF;
			/*FALLTHRU*/
		    }
		}
		rowOffset += imagePtr->bytes_per_line;
	    }
	}
	break;

    case DirectColor:
	{
	    register int x, y;
	    register Pix32 *srcPtr;
	    register char *destPtr;
	    unsigned int pixel;
	    int rowOffset;
	    struct ColorTableRec *colorTabPtr;

	    /* Build a color table first */
	    colorTabPtr = Blt_DirectColorTable(interp, tkwin, image);

	    /*
	     * Compute the colormap locations directly from pixel RGB values.
	     */
	    srcPtr = Blt_ColorimageData(image);
	    rowOffset = 0;
	    for (y = 0; y < height; y++) {
		destPtr = imagePtr->data + rowOffset;
		for (x = 0; x < width; x++, srcPtr++) {
		    pixel = DirectColorPixel(colorTabPtr, srcPtr);
		    switch (imagePtr->bits_per_pixel) {
		    case 32:
			*destPtr++ = (pixel >> 24) & 0xFF;
			/*FALLTHRU*/
		    case 24:
			*destPtr++ = (pixel >> 16) & 0xFF;
			/*FALLTHRU*/
		    case 16:
			*destPtr++ = (pixel >> 8) & 0xFF;
			/*FALLTHRU*/
		    case 8:
			*destPtr++ = pixel & 0xFF;
			/*FALLTHRU*/
		    }
		}
		rowOffset += imagePtr->bytes_per_line;
	    }
	    *colorTablePtr = colorTabPtr;
	}
	break;

    case GrayScale:
    case StaticGray:
    case PseudoColor:
    case StaticColor:
	{
	    register int x, y;
	    register Pix32 *srcPtr;
	    register char *destPtr;
	    unsigned int pixel;
	    int rowOffset;
	    struct ColorTableRec *colorTabPtr;

	    colorTabPtr = Blt_PseudoColorTable(interp, tkwin, image);

	    srcPtr = Blt_ColorimageData(image);
	    rowOffset = 0;
	    for (y = 0; y < height; y++) {
		destPtr = imagePtr->data + rowOffset;
		for (x = 0; x < width; x++, srcPtr++) {
		    pixel = PseudoColorPixel(srcPtr, colorTabPtr->lut);
		    switch (imagePtr->bits_per_pixel) {
		    case 32:
			*destPtr++ = (pixel >> 24) & 0xFF;
			/*FALLTHRU*/
		    case 24:
			*destPtr++ = (pixel >> 16) & 0xFF;
			/*FALLTHRU*/
		    case 16:
			*destPtr++ = (pixel >> 8) & 0xFF;
			/*FALLTHRU*/
		    case 8:
			*destPtr++ = pixel & 0xFF;
			/*FALLTHRU*/
		    }
		}
		rowOffset += imagePtr->bytes_per_line;
	    }
	    Blt_Free((char *)colorTabPtr->lut);
	    *colorTablePtr = colorTabPtr;
	}
	break;
    default:
	return None;		/* Bad or unknown visual class. */
    }
    pixmapGC = Tk_GetGC(tkwin, 0L, (XGCValues *)NULL);
    pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height,
	Tk_Depth(tkwin));
    XPutImage(display, pixmap, pixmapGC, imagePtr, 0, 0, 0, 0, width, height);
    XDestroyImage(imagePtr);
    Tk_FreeGC(display, pixmapGC);
    return pixmap;
}

#endif /* WIN32 */
/*
 *	filter function definitions
 */

#ifdef __STDC__
static Filter1DProc DefaultFilter;
static Filter1DProc BellFilter;
static Filter1DProc BesselFilter;
static Filter1DProc BoxFilter;
static Filter1DProc BSplineFilter;
static Filter1DProc CatRomFilter;
static Filter1DProc DummyFilter;
static Filter1DProc GaussianFilter;
static Filter1DProc Lanczos3Filter;
static Filter1DProc MitchellFilter;
static Filter1DProc SincFilter;
static Filter1DProc TriangleFilter;
static Tk_ImageChangedProc TempImageChangedProc;
#endif

static double
DefaultFilter(x)
    double x;
{
    if (x < 0.0) {
	x = -x;
    }
    if (x < 1.0) {
	/* f(x) = 2x^3 - 3x^2 + 1, -1 <= x <= 1 */
	return (2.0 * x - 3.0) * x * x + 1.0;
    }
    return 0.0;
}

/* Just for testing */
static double
DummyFilter(x)
    double x;
{
    return FABS(x);
}

/*
 *
 * Finite filters in increasing order:
 *	Box (constant)
 *	Triangle (linear)
 *	Bell
 *	BSpline (cubic)
 *
 */
static double
BoxFilter(x)
    double x;
{
    if (x < -0.5) {
	return 0.0;
    }
    if (x <= 0.5) {
	return 1.0;
    }
    return 0.0;
}

static double
TriangleFilter(x)
    double x;
{
    if (x < 0.0) {
	x = -x;
    }
    if (x < 1.0) {
	return (1.0 - x);
    }
    return 0.0;
}

static double
BellFilter(x)
    double x;
{
    if (x < 0) {
	x = -x;
    }
    if (x < 0.5) {
	return (0.75 - (x * x));
    }
    if (x < 1.5) {
	x = (x - 1.5);
	return (0.5 * (x * x));
    }
    return 0.0;
}

static double
BSplineFilter(x)
    double x;
{
    double x2;

    if (x < 0) {
	x = -x;
    }
    if (x < 1) {
	x2 = x * x;
	return ((.5 * x2 * x) - x2 + (2.0 / 3.0));
    } else if (x < 2) {
	x = 2 - x;
	return ((x * x * x) / 6.0);
    }
    return 0.0;
}

/*
 *
 * Infinite Filters:
 *	Sinc		perfect lowpass filter
 *	Bessel		circularly symmetric 2-D filter
 *	Gaussian
 *	Lanczos3
 *	Mitchell
 */

static double
SincFilter(x)
    double x;
{
    if (x == 0.0) {
	return 1.0;
    }
    x *= M_PI;
    return (sin(x) / x);
}

static double
BesselFilter(x)
    double x;
{
#ifdef NEED_DECL_J1
    extern double j1 _ANSI_ARGS_((double value));
#endif
    /*
     * See Pratt "Digital Image Processing" p. 97 for Bessel functions
     * zeros are at approx x=1.2197, 2.2331, 3.2383, 4.2411, 5.2428, 6.2439,
     * 7.2448, 8.2454
     */
    return (x == 0.0) ? M_PI / 4.0 : j1(M_PI * x) / (x + x);
}

#define SQRT_2PI	0.79788456080286541	/* sqrt(2.0 / M_PI) */

static double
GaussianFilter(x)
    double x;
{
    return exp(-2.0 * x * x) * SQRT_2PI;
}

static double
Lanczos3Filter(x)
    double x;
{
    if (x < 0) {
	x = -x;
    }
    if (x < 3.0) {
	return (SincFilter(x) * SincFilter(x / 3.0));
    }
    return 0.0;
}

#define	B		0.3333333333333333	/* (1.0 / 3.0) */
#define	C		0.3333333333333333	/* (1.0 / 3.0) */

static double
MitchellFilter(x)
    double x;
{
    double x2;

    x2 = x * x;
    if (x < 0) {
	x = -x;
    }
    if (x < 1.0) {
	x = (((12.0 - 9.0 * B - 6.0 * C) * (x * x2)) +
	    ((-18.0 + 12.0 * B + 6.0 * C) * x2) + (6.0 - 2 * B));
	return (x / 6.0);
    } else if (x < 2.0) {
	x = (((-1.0 * B - 6.0 * C) * (x * x2)) + ((6.0 * B + 30.0 * C) * x2) +
	    ((-12.0 * B - 48.0 * C) * x) + (8.0 * B + 24 * C));
	return (x / 6.0);
    }
    return 0.0;
}

/*
 * Catmull-Rom spline
 */
static double
CatRomFilter(x)
    double x;
{
    if (x < -2.) {
	return 0.0;
    }
    if (x < -1.0) {
	return 0.5 * (4.0 + x * (8.0 + x * (5.0 + x)));
    }
    if (x < 0.0) {
	return 0.5 * (2.0 + x * x * (-5.0 + x * -3.0));
    }
    if (x < 1.0) {
	return 0.5 * (2.0 + x * x * (-5.0 + x * 3.0));
    }
    if (x < 2.0) {
	return 0.5 * (4.0 + x * (-8.0 + x * (5.0 - x)));
    }
    return 0.0;
}

static Filter1D filterTable[] =
{
    /* name,     function,		support */
    {"bell",     BellFilter,		1.5	 },
    {"bessel",   BesselFilter,		3.2383   },
    {"box",      BoxFilter,		0.5      },
    {"bspline",  BSplineFilter,		2.0	 },
    {"catrom",   CatRomFilter,		2.0	 },
    {"default",  DefaultFilter,		1.0	 },
    {"dummy",    DummyFilter,		0.5	 },
    {"gauss8",   GaussianFilter,	8.0	 },
    {"gaussian", GaussianFilter,	4.0	 },
    {"lanczos3", Lanczos3Filter,	3.0	 },
    {"mitchell", MitchellFilter,	2.0	 },
    {"none",     (Filter1DProc *)NULL,	0.0	 },
    {"sinc",     SincFilter,		2.0	 },
    {"triangle", TriangleFilter,	1.0	 },
};

static int nFilters = sizeof(filterTable) / sizeof(Filter1D);

Filter1D *bltBoxFilterPtr = &(filterTable[1]);


/*
 *----------------------------------------------------------------------
 *
 * Blt_Get1DFilter --
 *
 *      Finds a 1-D filter associated by the given filter name.
 *
 * Results:
 *      A standard Tcl result.  Returns TCL_OK is the filter was
 *	found.  The filter information (proc and support) is returned
 *	via filterPtrPtr. Otherwise TCL_ERROR is returned and an error
 *	message is left in interp->result.
 *
 *----------------------------------------------------------------------
 */
int
Blt_Get1DFilter(interp, name, filterPtrPtr)
    Tcl_Interp *interp;
    char *name;
    Filter1D **filterPtrPtr;
{
    Filter1D *filterPtr, *endPtr;

    endPtr = filterTable + nFilters;
    for (filterPtr = filterTable; filterPtr < endPtr; filterPtr++) {
	if (strcmp(name, filterPtr->name) == 0) {
	    *filterPtrPtr = (filterPtr->proc == NULL) ? NULL : filterPtr;
	    return TCL_OK;
	}
    }
    Tcl_AppendResult(interp, "can't find filter \"", name, "\"", (char *)NULL);
    return TCL_ERROR;
}


typedef struct {
    int pixel;
    float weight;
} Sample;

typedef struct {
    int count;			/* Number of contributors */
    Sample *sampleArr;		/* Array of contributors */
} Contribution;

static void
ComputeContributors(srcWidth, destWidth, filterPtr, contribPtrPtr,
    samplePtrPtr)
    int srcWidth, destWidth;
    Filter1D *filterPtr;
    Contribution **contribPtrPtr;
    Sample **samplePtrPtr;
{
    Contribution *contribArr;
    Sample *samplePool, *sampleArr;
    double scale;
    int sampleSize;
    double center;
    register Contribution *c;
    register Sample *s;
    register int x, i;
    register int left, right;	/* filter bounds */
    float factor, sum;

    /* Pre-calculate filter contributions for a row */
    contribArr = (Contribution *)Blt_Calloc(destWidth, sizeof(Contribution));
    assert(contribArr);

    scale = (double)destWidth / (double)srcWidth;

    if (scale < 1.0) {
	double width, fscale;

	/* Downsample */

	width = filterPtr->support / scale;
	fscale = 1.0 / scale;
	sampleSize = (int)(width * 2 + 2);
	samplePool = (Sample *)
	    Blt_Malloc(sampleSize * destWidth * sizeof(Sample));
	sampleArr = samplePool;
	for (c = contribArr, x = 0; x < destWidth; x++, c++) {
	    center = (double)x *fscale;

	    /* Determine bounds of filter and its density */
	    left = (int)(center - width + 0.5);
	    if (left < 0) {
		left = 0;
	    }
	    right = (int)(center + width + 0.5);
	    if (right >= srcWidth) {
		right = srcWidth - 1;
	    }
	    sum = 0.0;
	    for (s = sampleArr, i = left; i <= right; i++, s++) {
		s->weight = (float)
		    (*filterPtr->proc) (((double)i + 0.5 - center) * scale);
		sum += s->weight;
	    }
	    factor = (float)((sum == 0.0) ? 1.0 : (1.0 / sum));
	    for (s = sampleArr, i = left; i <= right; i++, s++) {
		s->pixel = i;
		s->weight *= factor;
	    }
	    c->count = right - left + 1;
	    c->sampleArr = sampleArr;
	    sampleArr += sampleSize;
	}
    } else {
	/* Upsample */

	sampleSize = (int)(filterPtr->support * 2 + 2);
	samplePool = (Sample *)
	    Blt_Malloc(sampleSize * destWidth * sizeof(Sample));
	sampleArr = samplePool;

	for (c = contribArr, x = 0; x < destWidth; x++, c++) {
	    center = (double)x / scale;
	    left = (int)(center - filterPtr->support + 0.5);
	    if (left < 0) {
		left = 0;
	    }
	    right = (int)(center + filterPtr->support + 0.5);
	    if (right >= srcWidth) {
		right = srcWidth - 1;
	    }
	    sum = 0.0;
	    for (s = sampleArr, i = left; i <= right; i++, s++) {
		s->weight =
		    (float)(*filterPtr->proc) ((double)i + 0.5 - center);
		sum += s->weight;
	    }
	    factor = (float)((sum == 0.0) ? 1.0 : (1.0 / sum));
	    for (s = sampleArr, i = left; i <= right;
		i++, s++) {
		s->pixel = i;
		s->weight *= factor;
	    }
	    c->count = right - left + 1;
	    c->sampleArr = sampleArr;
	    sampleArr += sampleSize;
	}
    }
    *samplePtrPtr = samplePool;
    *contribPtrPtr = contribArr;
}

static Blt_Colorimage
ZoomImageVertically(src, srcRegionPtr, destRegionPtr, filterPtr)
    Blt_Colorimage src;
    Region2D *srcRegionPtr;	/* Region of source image to resample */
    Region2D *destRegionPtr;	/* Region of destination image to redraw */
    Filter1D *filterPtr;
{
    Contribution *contribArr, *c;
    register Sample *s, *endPtr;
    Sample *sampleArr;
    register int x, y;
    Blt_Colorimage dest;
    register Pix32 *srcPtr, *destPtr;
    float red, green, blue;
    Pix32 *pixelPtr;
    int destOffset;
    int srcWidth, srcHeight;
    int destWidth, destHeight;

    srcWidth = RegionWidth(srcRegionPtr);
    srcHeight = RegionHeight(srcRegionPtr);
    destWidth = RegionWidth(destRegionPtr);
    destHeight = RegionHeight(destRegionPtr);

    dest = Blt_CreateColorimage(destWidth, destHeight);

    /* Pre-calculate filter contributions for a row */
    ComputeContributors(srcHeight, destHeight, filterPtr, &contribArr, 
	&sampleArr);
    /* Apply filter to zoom vertically from tmp to destination */
    for (x = 0; x < srcWidth; x++) {
	destOffset = x;
	srcPtr = Blt_ColorimageData(src) + x;
	for (c = contribArr, y = 0; y < destHeight; y++, c++) {
	    red = green = blue = 0.0;
	    endPtr = c->sampleArr + c->count;
	    for (s = c->sampleArr; s < endPtr; s++) {
		pixelPtr = srcPtr + (s->pixel * srcWidth);
		red += (float)pixelPtr->Red * s->weight;
		green += (float)pixelPtr->Green * s->weight;
		blue += (float)pixelPtr->Blue * s->weight;
	    }
	    destPtr = Blt_ColorimageData(dest) + destOffset;
	    destPtr->Red = (unsigned char)CLAMP(red);
	    destPtr->Green = (unsigned char)CLAMP(green);
	    destPtr->Blue = (unsigned char)CLAMP(blue);
	    destPtr->Alpha = (unsigned char)-1;
	    destOffset += destWidth;
	}
    }

    /* Free the memory allocated for filter weights */
    Blt_Free((char *)sampleArr);
    Blt_Free((char *)contribArr);

    return dest;
}

static Blt_Colorimage
ZoomImageHorizontally(src, srcRegionPtr, destRegionPtr, filterPtr)
    Blt_Colorimage src;
    Region2D *srcRegionPtr;	/* Region of source image to resample */
    Region2D *destRegionPtr;	/* Region of destination image to redraw */
    Filter1D *filterPtr;
{
    Contribution *contribArr, *c;
    register Sample *s, *endPtr;
    Sample *sampleArr;
    register int x, y;
    Blt_Colorimage dest;
    register Pix32 *srcPtr, *destPtr;
    float red, green, blue;
    Pix32 *pixelPtr;
    int srcOffset, destOffset;
    int srcWidth, srcHeight;
    int destWidth, destHeight;

    srcWidth = RegionWidth(srcRegionPtr);
    srcHeight = RegionHeight(srcRegionPtr);
    destWidth = RegionWidth(destRegionPtr);
    destHeight = RegionHeight(destRegionPtr);

    dest = Blt_CreateColorimage(destWidth, destHeight);

    /* Pre-calculate filter contributions for a row */
    ComputeContributors(srcWidth, destWidth, filterPtr, &contribArr, 
	&sampleArr);

    /* Apply filter to zoom horizontally from srcPtr to tmpPixels */
    srcOffset = srcRegionPtr->left + (srcRegionPtr->top * srcWidth);
    destOffset = 0;
    for (y = 0; y < srcHeight; y++) {
	srcPtr = Blt_ColorimageData(src) + srcOffset;
	destPtr = Blt_ColorimageData(dest) + destOffset;
	for (c = contribArr, x = 0; x < destWidth; x++, c++) {
	    red = green = blue = 0.0;
	    endPtr = c->sampleArr + c->count;
	    for (s = c->sampleArr; s < endPtr; s++) {
		pixelPtr = srcPtr + s->pixel;
		red += (float)pixelPtr->Red * s->weight;
		green += (float)pixelPtr->Green * s->weight;
		blue += (float)pixelPtr->Blue * s->weight;
	    }
	    destPtr->Red = (unsigned char)CLAMP(red);
	    destPtr->Green = (unsigned char)CLAMP(green);
	    destPtr->Blue = (unsigned char)CLAMP(blue);
	    destPtr->Alpha = (unsigned char)-1;
	    destPtr++;
	}
	destOffset += destWidth;
	srcOffset += srcWidth;
    }
    /* free the memory allocated for horizontal filter weights */
    Blt_Free((char *)sampleArr);
    Blt_Free((char *)contribArr);

    return dest;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ResampleColorimage --
 *
 *      Resamples a given color image using 1-D filters and returns
 *	a new color image of the designated size.
 *
 * Results:
 *      Returns the resampled color image. The original color image
 *	is left intact.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_ResampleColorimage(src, srcRegionPtr, destRegionPtr, horzFilterPtr,
    vertFilterPtr)
    Blt_Colorimage src;
    Region2D *srcRegionPtr;	/* Region of source image to resample */
    Region2D *destRegionPtr;	/* Region of destination image to redraw */
    Filter1D *horzFilterPtr, *vertFilterPtr;
{
    Blt_Colorimage tmp, dest;
    Region2D tmpRegion, srcRegion;
    int count1, count2;
    int vertSupport, horzSupport;

    vertSupport = (int)ceil(vertFilterPtr->support * 2.0);
    horzSupport = (int)ceil(horzFilterPtr->support * 2.0);

    if (srcRegionPtr == NULL) {
	srcRegionPtr = ColorimageRegion(src, &srcRegion);
    }
    tmpRegion.left = tmpRegion.top = 0;

    /* Determine if it's cheaper to first zoom vertically or horizontally */
    count1 = RegionWidth(destRegionPtr) * (RegionHeight(srcRegionPtr) *
	horzSupport + RegionHeight(destRegionPtr) * vertSupport);
    count2 = RegionHeight(destRegionPtr) * (RegionWidth(srcRegionPtr) *
	vertSupport + RegionWidth(destRegionPtr) * horzSupport);
    if (count1 < count2) {
	tmpRegion.right = srcRegionPtr->right;
	tmpRegion.bottom = destRegionPtr->bottom;
	tmp = ZoomImageVertically(src, srcRegionPtr, &tmpRegion, vertFilterPtr);
	dest = ZoomImageHorizontally(tmp, &tmpRegion, destRegionPtr,
	    horzFilterPtr);
    } else {
	tmpRegion.bottom = srcRegionPtr->bottom;
	tmpRegion.right = destRegionPtr->right;
	tmp = ZoomImageHorizontally(src, srcRegionPtr, &tmpRegion,
		horzFilterPtr);
	dest = ZoomImageVertically(tmp, &tmpRegion, destRegionPtr, 
		vertFilterPtr);
    }
    Blt_FreeColorimage(tmp);
    return dest;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ResamplePhoto --
 *
 *      Resamples a Tk photo image using 1-D filters and writes the
 *      image into another Tk photo.  It is possible for the
 *      source and destination to be the same photo.
 *
 * Results:
 *      The designated destination photo will contain the resampled
 *	color image. The original photo is left intact.
 *
 *---------------------------------------------------------------------- 
 */
void
Blt_ResamplePhoto(srcPhoto, destPhoto, srcRegionPtr, horzFilterPtr, 
	vertFilterPtr)
    Tk_PhotoHandle srcPhoto;	/* Source photo image to scale */
    Tk_PhotoHandle destPhoto;	/* Resulting scaled photo image */
    Region2D *srcRegionPtr;	/* Region of source image to resample */
    Filter1D *horzFilterPtr, *vertFilterPtr;
{
    Blt_Colorimage srcImage, destImage;
    Tk_PhotoImageBlock src, dest;
    Region2D srcRegion, destRegion;

    Tk_PhotoGetImage(srcPhoto, &src);
    Tk_PhotoGetImage(destPhoto, &dest);

    if (srcRegionPtr == NULL) {
	srcRegionPtr = Blt_SetRegion(0, 0, src.width, src.height, &srcRegion);
    }
    srcImage = Blt_PhotoToColorimage(srcPhoto, srcRegionPtr);
    Blt_SetRegion(0, 0, dest.width, dest.height, &destRegion);
    destImage = Blt_ResampleColorimage(srcImage, NULL, &destRegion, 
	horzFilterPtr, vertFilterPtr);
    Blt_FreeColorimage(srcImage);
    Blt_ColorimageToPhoto(destImage, destPhoto);
    Blt_FreeColorimage(destImage);
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ResizePhoto --
 *
 *	Scales the region of the source image to the size of the
 *	destination image.  This routine performs raw scaling of
 *	the image and unlike Blt_ResamplePhoto does not handle
 *	aliasing effects from subpixel sampling. It is possible
 *	for the source and destination to be the same photo.
 *
 * Results:
 *      The designated destination photo will contain the resampled
 *	color image. The original photo is left intact.
 *
 *----------------------------------------------------------------------
 */
void
Blt_ResizePhoto(srcPhoto, destPhoto, regionPtr)
    Tk_PhotoHandle srcPhoto;	/* Source photo image to scale */
    Tk_PhotoHandle destPhoto;	/* Resulting scaled photo image */
    Region2D *regionPtr;	/* Region of source image to be scaled. 
				 * If NULL, use the entire image. */
{
    register int x, y;
    register int sx, sy;
    float xScale, yScale;
    Region2D region;
    Blt_Colorimage destImage;
    Pix32 *destPtr;
    Tk_PhotoImageBlock src, dest;
    unsigned char *srcPtr;
    int offset;

    Tk_PhotoGetImage(srcPhoto, &src);
    Tk_PhotoGetImage(destPhoto, &dest);

    if (regionPtr == NULL) {
	regionPtr = Blt_SetRegion(0, 0, src.width, src.height, &region);
    }
    destImage = Blt_CreateColorimage(dest.width, dest.height);
    xScale = (float)RegionWidth(regionPtr) / (float)dest.width;
    yScale = (float)RegionHeight(regionPtr) / (float)dest.height;

    destPtr = Blt_ColorimageData(destImage);
    if (src.pixelSize == 4) {
	for (y = 0; y < dest.height; y++) {
	    sy = (int)(yScale * (float)y) + regionPtr->top;
	    if (sy > regionPtr->bottom) {
		sy = regionPtr->bottom;
	    }
	    offset = (sy * src.pitch);
	    for (x = 0; x < dest.width; x++) {
		sx = (int)(xScale * (float)x) + regionPtr->left;
		if (sx > regionPtr->right) {
		    sx = regionPtr->right;
		}
		srcPtr = src.pixelPtr + offset + (sx * src.pixelSize);
		destPtr->Red = srcPtr[src.offset[0]];
		destPtr->Green = srcPtr[src.offset[1]];
		destPtr->Blue = srcPtr[src.offset[2]];
		destPtr->Alpha = srcPtr[src.offset[3]];
		destPtr++;
	    }
	}
    } else if (src.pixelSize == 3) {
	for (y = 0; y < dest.height; y++) {
	    sy = (int)(yScale * (float)y) + regionPtr->top;
	    if (sy > regionPtr->bottom) {
		sy = regionPtr->bottom;
	    }
	    offset = (sy * src.pitch);
	    for (x = 0; x < dest.width; x++) {
		sx = (int)(xScale * (float)x) + regionPtr->left;
		if (sx > regionPtr->right) {
		    sx = regionPtr->right;
		}
		srcPtr = src.pixelPtr + offset + (sx * src.pixelSize);
		destPtr->Red = srcPtr[src.offset[0]];
		destPtr->Green = srcPtr[src.offset[1]];
		destPtr->Blue = srcPtr[src.offset[2]];
		destPtr->Alpha = (unsigned char)-1;
		destPtr++;
	    }
	}
    } else  {
	for (y = 0; y < dest.height; y++) {
	    sy = (int)(yScale * (float)y) + regionPtr->top;
	    if (sy > regionPtr->bottom) {
		sy = regionPtr->bottom;
	    }
	    offset = (sy * src.pitch);
	    for (x = 0; x < dest.width; x++) {
		sx = (int)(xScale * (float)x) + regionPtr->left;
		if (sx > regionPtr->right) {
		    sx = regionPtr->right;
		}
		srcPtr = src.pixelPtr + offset + (sx * src.pixelSize);
		destPtr->Red = destPtr->Green = destPtr->Blue = 
		    srcPtr[src.offset[0]];
		destPtr->Alpha = (unsigned char)-1;
		destPtr++;
	    }
	}
    }	
    Blt_ColorimageToPhoto(destImage, destPhoto);
    Blt_FreeColorimage(destImage);
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_ResizeColorimage --
 *
 *	Scales the region of the source image to the size of the
 *	destination image.  This routine performs raw scaling of
 *	the image and unlike Blt_ResamplePhoto does not perform
 *	any antialiasing.
 *
 * Results:
 *      Returns the new resized color image.  The original image
 *	is left intact.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_ResizeColorimage(src, regionPtr, destWidth, destHeight)
    Blt_Colorimage src;		/* Source photo image to scale */
    Region2D *regionPtr;	/* Region of source image to be scaled.
				 * If NULL, use the entire image. */
    int destWidth, destHeight;	/* Dimensions of resized image */
{
    register int x, y;
    register int sx, sy;
    float xScale, yScale;
    Blt_Colorimage dest;
    Region2D region;
    Pix32 *srcPtr, *destPtr;
    int width, height;

    dest = Blt_CreateColorimage(destWidth, destHeight);
    if (regionPtr == NULL) {
	regionPtr = ColorimageRegion(src, &region);
    }
    width = RegionWidth(regionPtr);
    height = RegionHeight(regionPtr);
    xScale = (float)width / (float)destWidth;
    yScale = (float)height / (float)destHeight;

    destPtr = Blt_ColorimageData(dest);
    for (y = 0; y < destHeight; y++) {
	sy = (int)(yScale * (float)y) + regionPtr->top;
	if (sy > regionPtr->bottom) {
	    sy = regionPtr->bottom;
	} 
	for (x = 0; x < destWidth; x++) {
	    sx = (int)(xScale * (float)x) + regionPtr->left;
	    if (sx > regionPtr->right) {
		sx = regionPtr->right;
	    } 
	    srcPtr = Blt_ColorimagePixel(src, sx, sy);
	    destPtr->Red = srcPtr->Red;
	    destPtr->Green = srcPtr->Green;
	    destPtr->Blue = srcPtr->Blue;
	    destPtr->Alpha = srcPtr->Alpha;
	    destPtr++;
	}
    }
    return dest;
}

/*
 * FIXME: Boundary handling could be better (pixels are replicated).
 *	  It's slow. Take boundary tests out of inner loop.
 */
Blt_Colorimage
Blt_ConvolveColorimage(src, filterPtr)
    Blt_Colorimage src;
    Filter2D *filterPtr;
{
    Blt_Colorimage dest;
    register Pix32 *srcPtr, *destPtr;
#define MAXROWS	24
    register int sx, sy, dx, dy;
    register int x, y;
    float red, green, blue;
    int width, height;
    int radius;
    register float *valuePtr;

    width = Blt_ColorimageWidth(src);
    height = Blt_ColorimageHeight(src);

    dest = Blt_CreateColorimage(width, height);
    radius = (int)filterPtr->support;
    if (radius < 1) {
	radius = 1;
    }
    destPtr = Blt_ColorimageData(dest);
    for (dy = 0; dy < height; dy++) {
	for (dx = 0; dx < width; dx++) {
	    red = green = blue = 0.0;
	    valuePtr = filterPtr->kernel;
	    for (sy = (dy - radius); sy <= (dy + radius); sy++) {
		y = sy;
		if (y < 0) {
		    y = 0;
		} else if (y >= height) {
		    y = height - 1;
		}
		for (sx = (dx - radius); sx <= (dx + radius); sx++) {
		    x = sx;
		    if (x < 0) {
			x = 0;
		    } else if (sx >= width) {
			x = width - 1;
		    }
		    srcPtr = Blt_ColorimagePixel(src, x, y);
		    red += *valuePtr * (float)srcPtr->Red;
		    green += *valuePtr * (float)srcPtr->Green;
		    blue += *valuePtr * (float)srcPtr->Blue;
#ifdef notdef
		    fprintf(stderr, "%d,%d = r=%f,g=%f,b=%f\n", x, y,
			red, green, blue);
#endif
		    valuePtr++;
		}
	    }
	    red /= filterPtr->sum;
	    green /= filterPtr->sum;
	    blue /= filterPtr->sum;
	    destPtr->Red = (unsigned char)CLAMP(red);
	    destPtr->Green = (unsigned char)CLAMP(green);
	    destPtr->Blue = (unsigned char)CLAMP(blue);
	    destPtr->Alpha = (unsigned char)-1;
	    destPtr++;
	}
    }
    return dest;
}

#ifdef WIN32

/*
 *----------------------------------------------------------------------
 *
 * Blt_DrawableRegionToColorimage --
 *
 *      Takes a snapshot of an X drawable (pixmap or window) and
 *	converts it to a color image.
 *
 * Results:
 *      Returns a color image of the drawable.  If an error occurred,
 *	NULL is returned.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_DrawableRegionToColorimage(tkwin, drawable, width, height, regionPtr)
    Tk_Window tkwin;
    Drawable drawable;
    int width, height;		/* Dimension of the drawable. */
    Region2D *regionPtr;	/* Region to be snapped. */
{
    void *data;
    BITMAPINFO info;
    DIBSECTION ds;
    HBITMAP hBitmap, oldBitmap;
    HPALETTE hPalette;
    HDC memDC;
    unsigned char *srcArr;
    register unsigned char *srcPtr;
    HDC hDC;
    TkWinDCState state;
    register Pix32 *destPtr;
    Blt_Colorimage image;
    register int x, y;

    hDC = TkWinGetDrawableDC(Tk_Display(tkwin), drawable, &state);

    /* Create the intermediate drawing surface at window resolution. */
    ZeroMemory(&info, sizeof(info));
    info.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    info.bmiHeader.biWidth = width;
    info.bmiHeader.biHeight = height;
    info.bmiHeader.biPlanes = 1;
    info.bmiHeader.biBitCount = 32;
    info.bmiHeader.biCompression = BI_RGB;
    hBitmap = CreateDIBSection(hDC, &info, DIB_RGB_COLORS, &data, NULL, 0);
    memDC = CreateCompatibleDC(hDC);
    oldBitmap = SelectBitmap(memDC, hBitmap);

    hPalette = Blt_GetSystemPalette();
    if (hPalette != NULL) {
	SelectPalette(hDC, hPalette, FALSE);
	RealizePalette(hDC);
	SelectPalette(memDC, hPalette, FALSE);
	RealizePalette(memDC);
    }
    image = NULL;
    /* Copy the window contents to the memory surface. */
    if (!BitBlt(memDC, 0, 0, width, height, hDC, regionPtr->left, 
	regionPtr->top, SRCCOPY)) {
#ifdef notdef
	PurifyPrintf("can't blit: %s\n", Blt_LastError());
#endif
	goto done;
    }
    if (GetObject(hBitmap, sizeof(DIBSECTION), &ds) == 0) {
#ifdef notdef
	PurifyPrintf("can't get object: %s\n", Blt_LastError());
#endif
	goto done;
    }
    srcArr = (unsigned char *)ds.dsBm.bmBits;
    image = Blt_CreateColorimage(width, height);
    destPtr = Blt_ColorimageData(image);

    /* 
     * Copy the DIB RGB data into the color image. The DIB scanlines
     * are stored bottom-to-top and the order of the RGB color
     * components is BGR. Who says Win32 GDI programming isn't
     * backwards?  
     */
    for (y = height - 1; y >= 0; y--) {
	srcPtr = srcArr + (y * ds.dsBm.bmWidthBytes);
	for (x = 0; x < width; x++) {
	    destPtr->Blue = *srcPtr++;
	    destPtr->Green = *srcPtr++;
	    destPtr->Red = *srcPtr++;
	    destPtr->Alpha = (unsigned char)-1;
	    destPtr++;
	    srcPtr++;
	}
    }
  done:
    DeleteBitmap(SelectBitmap(memDC, oldBitmap));
    DeleteDC(memDC);
    TkWinReleaseDrawableDC(drawable, hDC, &state);
    if (hPalette != NULL) {
	DeletePalette(hPalette);
    }
    return image;
}

#else

/* ARGSUSED */
static int
XGetImageErrorProc(clientData, errEventPtr)
    ClientData clientData;
    XErrorEvent *errEventPtr;
{
    int *errorPtr = (int *)clientData;

    *errorPtr = TCL_ERROR;
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_DrawableRegionToColorimage --
 *
 *      Takes a snapshot of an X drawable (pixmap or window) and
 *	converts it to a color image.
 *
 *	The trick here is to efficiently convert the pixel values
 *	(indices into the color table) into RGB color values.  In the
 *	days of 8-bit displays, it was simpler to get RGB values for
 *	all 256 indices into the colormap.  Instead we'll build a
 *	hashtable of unique pixels and from that an array of pixels to
 *	pass to XQueryColors.  For TrueColor visuals, we'll simple
 *	compute the colors from the pixel.
 *
 *	[I don't know how much faster it would be to take advantage
 *	of all the different visual types.  This pretty much depends
 *	on the size of the image and the number of colors it uses.]
 *
 * Results:
 *      Returns a color image of the drawable.  If an error occurred,
 *	NULL is returned.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_DrawableRegionToColorimage(tkwin, drawable, width, height, regionPtr)
    Tk_Window tkwin;
    Drawable drawable;
    int width, height;		/* Dimension of the drawable. */
    Region2D *regionPtr;	/* Region to be snapped. */
{
    XImage *imagePtr;
    Blt_Colorimage image;
    register Pix32 *destPtr;
    unsigned long pixel;
    int result = TCL_OK;
    Tk_ErrorHandler errHandler;
    register int x, y;
    Visual *visualPtr;

    errHandler = Tk_CreateErrorHandler(Tk_Display(tkwin), BadMatch,
	X_GetImage, -1, XGetImageErrorProc, (ClientData)&result);
    imagePtr = XGetImage(Tk_Display(tkwin), drawable, 0, 0, width, height,
	AllPlanes, ZPixmap);
    Tk_DeleteErrorHandler(errHandler);
    XSync(Tk_Display(tkwin), False);
    if (result != TCL_OK) {
	return NULL;
    }
    /*
     * First allocate an array of pixels big enough to hold the image.
     */
    image = Blt_CreateColorimage(RegionWidth(regionPtr), 
	RegionHeight(regionPtr));

    visualPtr = Tk_Visual(tkwin);
    if (visualPtr->class == TrueColor) {
	unsigned int red, green, blue;
	/*
	 * Directly compute the RGB color values from the pixel index
	 * rather than of going through XQueryColors.
	 */
	ComputeMasks(visualPtr);
	destPtr = Blt_ColorimageData(image);
	for (y = regionPtr->top; y <= regionPtr->bottom; y++) {
	    for (x = regionPtr->left; x <= regionPtr->right; x++) {
		pixel = XGetPixel(imagePtr, x, y);

		red = (pixel & visualPtr->red_mask) >> redMaskShift;
		green = (pixel & visualPtr->green_mask) >> greenMaskShift;
		blue = (pixel & visualPtr->blue_mask) >> blueMaskShift;

		/*
		 * The number of bits per color in the pixel may be
		 * less than eight. For example, 15/16 bit displays
		 * (hi-color) use only 5 bits, 8-bit displays use 2 or
		 * 3 bits (don't ask me why you'd have an 8-bit
		 * TrueColor display). So shift back the least
		 * significant bits.
		 */
		destPtr->Red = (unsigned char)(red << redAdjust);
		destPtr->Green = (unsigned char)(green << greenAdjust);
		destPtr->Blue = (unsigned char)(blue << blueAdjust);
		destPtr->Alpha = (unsigned char)-1;
		destPtr++;
	    }
	}
	XDestroyImage(imagePtr);
    } else {
	Blt_HashEntry *hPtr;
	Blt_HashSearch cursor;
	Blt_HashTable pixelTable;
	XColor *colorPtr, *colorArr;
	Pix32 *endPtr;
	int nPixels;
	int nColors;
	int isNew;

	/*
	 * Fill the array with each pixel of the image. At the same time, build
	 * up a hashtable of the pixels used.
	 */
	nPixels = RegionWidth(regionPtr) * RegionHeight(regionPtr);
	Blt_InitHashTable(&pixelTable, BLT_ONE_WORD_KEYS);
	destPtr = Blt_ColorimageData(image);
	for (y = regionPtr->top; y <= regionPtr->bottom; y++) {
	    for (x = regionPtr->left; x <= regionPtr->right; x++) {
		pixel = XGetPixel(imagePtr, x, y);
		hPtr = Blt_CreateHashEntry(&pixelTable, (char *)pixel, &isNew);
		if (isNew) {
		    Blt_SetHashValue(hPtr, (char *)pixel);
		}
		destPtr->value = pixel;
		destPtr++;
	    }
	}
	XDestroyImage(imagePtr);

	/* 
	 * Convert the hashtable of pixels into an array of XColors so
	 * that we can call XQueryColors with it. XQueryColors will
	 * convert the pixels into their RGB values.  
	 */
	nColors = pixelTable.numEntries;
	colorArr = (XColor *)Blt_Malloc(sizeof(XColor) * nColors);
	assert(colorArr);

	colorPtr = colorArr;
	for (hPtr = Blt_FirstHashEntry(&pixelTable, &cursor); hPtr != NULL;
	    hPtr = Blt_NextHashEntry(&cursor)) {
	    colorPtr->pixel = (unsigned long)Blt_GetHashValue(hPtr);
	    Blt_SetHashValue(hPtr, (char *)colorPtr);
	    colorPtr++;
	}
	XQueryColors(Tk_Display(tkwin), Tk_Colormap(tkwin), colorArr, nColors);

	/* 
	 * Go again through the array of pixels, replacing each pixel
	 * of the image with its RGB value.  
	 */
	destPtr = Blt_ColorimageData(image);
	endPtr = destPtr + nPixels;
	for (/* empty */; destPtr < endPtr; destPtr++) {
	    hPtr = Blt_FindHashEntry(&pixelTable, (char *)destPtr->value);
	    colorPtr = (XColor *)Blt_GetHashValue(hPtr);
	    destPtr->Red = (unsigned char)(colorPtr->red >> 8);
	    destPtr->Green = (unsigned char)(colorPtr->green >> 8);
	    destPtr->Blue = (unsigned char)(colorPtr->blue >> 8);
	    destPtr->Alpha = (unsigned char)-1;
	}
	Blt_Free((char *)colorArr);
	Blt_DeleteHashTable(&pixelTable);
    }
    return image;
}

#endif /* WIN32 */

Blt_Colorimage
Blt_DrawableToColorimage(tkwin, drawable, width, height)
    Tk_Window tkwin;
    Drawable drawable;
    int width, height;		/* Dimension of the drawable. */
{
    Region2D region;

    Blt_SetRegion(0, 0, width, height, &region);
    return Blt_DrawableRegionToColorimage(tkwin, drawable, width, height, 
		  &region);
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_SnapPhoto --
 *
 *      Takes a snapshot of an X drawable (pixmap or window) and
 *	writes it to an existing Tk photo image.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side Effects:
 *	The named Tk photo is updated with the snapshot.
 *
 *----------------------------------------------------------------------
 */
int
Blt_SnapPhoto(interp, tkwin, drawable, width, height, destWidth, destHeight,
    photoName)
    Tcl_Interp *interp;		/* Interpreter to report errors back to */
    Tk_Window tkwin;
    Drawable drawable;		/* Window or pixmap to be snapped */
    int width, height;		/* Dimension of the drawable */
    int destWidth, destHeight;	/* Desired size of the Tk photo */
    char *photoName;		/* Name of an existing Tk photo image. */
{
    Tk_PhotoHandle photo;	/* The photo image to write into. */
    Blt_Colorimage image;

    photo = Blt_FindPhoto(interp, photoName);
    if (photo == NULL) {
	Tcl_AppendResult(interp, "image \"", photoName, "\" doesn't",
	    " exist or is not a photo image", (char *)NULL);
	return TCL_ERROR;
    }
    image = Blt_DrawableToColorimage(tkwin, drawable, width, height);
    if (image == NULL) {
	Tcl_AppendResult(interp,
	    "can't grab window or pixmap (possibly obscured?)", (char *)NULL);
	return TCL_ERROR;	/* Can't grab window image */
    }
    if ((destWidth != width) || (destHeight != height)) {
	Region2D destRegion;
	Blt_Colorimage destImage;

	/*
	 * The requested size for the destination image is different than
	 * that of the source snapshot.  Resample the image as necessary.
	 * We'll use a cheap box filter. I'm assuming that the destination
	 * image will typically be smaller than the original.
	 */
	destRegion.left = destRegion.top = 0;
	destRegion.right = destWidth - 1;
	destRegion.bottom = destHeight - 1;
	destImage = Blt_ResampleColorimage(image, (Region2D *)NULL, &destRegion,
		bltBoxFilterPtr, bltBoxFilterPtr);
	Blt_FreeColorimage(image);
	image = destImage;
    }
    Blt_ColorimageToPhoto(image, photo);
    Blt_FreeColorimage(image);
    return TCL_OK;
}

#if HAVE_JPEGLIB_H


#undef HAVE_STDLIB_H
#undef EXTERN
#include "jpeglib.h"
#include <setjmp.h>

typedef struct {
    struct jpeg_error_mgr pub;	/* "public" fields */
    jmp_buf jmpBuf;
    Tcl_DString dString;
} ReaderHandler;

static void ErrorProc _ANSI_ARGS_((j_common_ptr jpegInfo));
static void MessageProc _ANSI_ARGS_((j_common_ptr jpegInfo));

/*
 * Here's the routine that will replace the standard error_exit method:
 */

static void
ErrorProc(jpgPtr)
    j_common_ptr jpgPtr;
{
    ReaderHandler *handlerPtr = (ReaderHandler *)jpgPtr->err;

    (*handlerPtr->pub.output_message) (jpgPtr);
    longjmp(handlerPtr->jmpBuf, 1);
}

static void
MessageProc(jpgPtr)
    j_common_ptr jpgPtr;
{
    ReaderHandler *handlerPtr = (ReaderHandler *)jpgPtr->err;
    char buffer[JMSG_LENGTH_MAX];

    /* Create the message and append it into the dynamic string. */
    (*handlerPtr->pub.format_message) (jpgPtr, buffer);
    Tcl_DStringAppend(&(handlerPtr->dString), " ", -1);
    Tcl_DStringAppend(&(handlerPtr->dString), buffer, -1);
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_JPEGToColorimage --
 *
 *      Reads a JPEG file and converts it into a color image.
 *
 * Results:
 *      The color image is returned.  If an error occured, such
 *	as the designated file could not be opened, NULL is returned.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_JPEGToColorimage(interp, fileName)
    Tcl_Interp *interp;
    char *fileName;
{
    struct jpeg_decompress_struct jpg;
    Blt_Colorimage image;
    unsigned int imageWidth, imageHeight;
    register Pix32 *destPtr;
    ReaderHandler handler;
    FILE *f;
    JSAMPLE **readBuffer;
    int row_stride;
    register int i;
    register JSAMPLE *bufPtr;

    f = fopen(fileName, "rb");
    if (f == NULL) {
	Tcl_AppendResult(interp, "can't open \"", fileName, "\":",
	    Tcl_PosixError(interp), (char *)NULL);
	return NULL;
    }
    image = NULL;

    /* Step 1: allocate and initialize JPEG decompression object */

    /* We set up the normal JPEG error routines, then override error_exit. */
    jpg.err = jpeg_std_error(&handler.pub);
    handler.pub.error_exit = ErrorProc;
    handler.pub.output_message = MessageProc;

    Tcl_DStringInit(&(handler.dString));
    Tcl_DStringAppend(&(handler.dString), "error reading \"", -1);
    Tcl_DStringAppend(&(handler.dString), fileName, -1);
    Tcl_DStringAppend(&(handler.dString), "\": ", -1);

    if (setjmp(handler.jmpBuf)) {
	jpeg_destroy_decompress(&jpg);
	fclose(f);
	Tcl_DStringResult(interp, &(handler.dString));
	return NULL;
    }
    jpeg_create_decompress(&jpg);
    jpeg_stdio_src(&jpg, f);

    jpeg_read_header(&jpg, TRUE);	/* Step 3: read file parameters */

    jpeg_start_decompress(&jpg);	/* Step 5: Start decompressor */
    imageWidth = jpg.output_width;
    imageHeight = jpg.output_height;
    if ((imageWidth < 1) || (imageHeight < 1)) {
	Tcl_AppendResult(interp, "bad JPEG image size", (char *)NULL);
	fclose(f);
	return NULL;
    }
    /* JSAMPLEs per row in output buffer */
    row_stride = imageWidth * jpg.output_components;

    /* Make a one-row-high sample array that will go away when done
     * with image */
    readBuffer = (*jpg.mem->alloc_sarray) ((j_common_ptr)&jpg, JPOOL_IMAGE, 
	row_stride, 1);
    image = Blt_CreateColorimage(imageWidth, imageHeight);
    destPtr = Blt_ColorimageData(image);

    if (jpg.output_components == 1) {
	while (jpg.output_scanline < imageHeight) {
	    jpeg_read_scanlines(&jpg, readBuffer, 1);
	    bufPtr = readBuffer[0];
	    for (i = 0; i < (int)imageWidth; i++) {
		destPtr->Red = destPtr->Green = destPtr->Blue = *bufPtr++;
		destPtr->Alpha = (unsigned char)-1;
		destPtr++;
	    }
	}
    } else {
	while (jpg.output_scanline < imageHeight) {
	    jpeg_read_scanlines(&jpg, readBuffer, 1);
	    bufPtr = readBuffer[0];
	    for (i = 0; i < (int)imageWidth; i++) {
		destPtr->Red = *bufPtr++;
		destPtr->Green = *bufPtr++;
		destPtr->Blue = *bufPtr++;
		destPtr->Alpha = (unsigned char)-1;
		destPtr++;
	    }
	}
    }
    jpeg_finish_decompress(&jpg);	/* We can ignore the return value
					 * since suspension is not
					 * possible with the stdio data
					 * source.  */
    jpeg_destroy_decompress(&jpg);


    /*  
     * After finish_decompress, we can close the input file.  Here we
     * postpone it until after no more JPEG errors are possible, so as
     * to simplify the setjmp error logic above.  (Actually, I don't
     * think that jpeg_destroy can do an error exit, but why assume
     * anything...)  
     */
    fclose(f);

    /* 
     * At this point you may want to check to see whether any corrupt-data
     * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
     */
    if (handler.pub.num_warnings > 0) {
	Tcl_SetErrorCode(interp, "IMAGE", "JPEG", 
		 Tcl_DStringValue(&(handler.dString)), (char *)NULL);
    } else {
	Tcl_SetErrorCode(interp, "NONE", (char *)NULL);
    }
    /*
     * We're ready to call the Tk_Photo routines. They'll take the RGB
     * array we've processed to build the Tk image of the JPEG.
     */
    Tcl_DStringFree(&(handler.dString));
    return image;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_JPEGToPhoto --
 *
 *      Reads a JPEG file and converts it into a Tk photo.
 *
 * Results:
 *      A standard Tcl result.  If successful, TCL_OK is returned
 *	and the designated photo is re-written with the image.
 *	Otherwise, TCL_ERROR is returned and interp->result will
 *	contain an error message.
 *
 *----------------------------------------------------------------------
 */
int
Blt_JPEGToPhoto(interp, fileName, photo)
    Tcl_Interp *interp;
    char *fileName;
    Tk_PhotoHandle photo;	/* The photo image to write into. */
{
    Blt_Colorimage image;

    image = Blt_JPEGToColorimage(interp, fileName);
    if (image == NULL) {
	return TCL_ERROR;
    }
    Blt_ColorimageToPhoto(image, photo);
    Blt_FreeColorimage(image);
    return TCL_OK;
}
#endif /* HAVE_JPEGLIB_H */

/*
 *----------------------------------------------------------------------
 *
 * Blt_ColorimageToGreyscale --
 *
 *	Converts a color image to PostScript grey scale (1 component)
 *	output.  Luminosity isn't computed using the old NTSC formula,
 *
 *	    Y = 0.299 * Red + 0.587 * Green + 0.114 * Blue
 *
 *      but the following
 *
 *	    Y = 0.212671 * Red + 0.715160 * Green + 0.072169 * Blue
 *
 *	which better represents contemporary monitors.
 *
 * Results:
 *	The color image is converted to greyscale.
 *
 *----------------------------------------------------------------------
 */
void
Blt_ColorimageToGreyscale(image)
    Blt_Colorimage image;
{
    register Pix32 *dataPtr, *endPtr;
    float Y;
    int nPixels;
    int width, height;

    width = Blt_ColorimageWidth(image);
    height = Blt_ColorimageHeight(image);
    nPixels = width * height;
    dataPtr = Blt_ColorimageData(image);
    endPtr = dataPtr + nPixels;
    for (/* empty */; dataPtr < endPtr; dataPtr++) {
	Y = (float)((0.212671 * (float)dataPtr->Red) +
		    (0.715160 * (float)dataPtr->Green) +
		    (0.072169 * (float)dataPtr->Blue));
	dataPtr->Red = dataPtr->Green = dataPtr->Blue = (unsigned char)CLAMP(Y);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TempImageChangedProc
 *
 *	The image is over-written each time it's resized.  We always
 *	resample from the color image we saved when the photo image
 *	was specified (-image option). So we only worry if the image
 *	is deleted.
 *
 * Results:
 *	None.
 *
 *---------------------------------------------------------------------- 
 */
/* ARGSUSED */
static void
TempImageChangedProc(clientData, x, y, width, height, imageWidth, imageHeight)
    ClientData clientData;
    int x, y, width, height;	/* Not used. */
    int imageWidth, imageHeight;/* Not used. */
{
#ifdef notdef
    fprintf(stderr, "should be redrawing temp image\n");
#endif
}

Tk_Image 
Blt_CreateTemporaryImage(interp, tkwin, clientData)
    Tcl_Interp *interp;
    Tk_Window tkwin;
    ClientData clientData;
{
    Tk_Image token;
    char *name;			/* Contains image name. */

    if (Tcl_Eval(interp, "image create photo") != TCL_OK) {
	return NULL;
    }
    name = Tcl_GetStringResult(interp);
    token = Tk_GetImage(interp, tkwin, name, TempImageChangedProc, clientData);
    if (token == NULL) {
	return NULL;
    }
    return token;
}

int
Blt_DestroyTemporaryImage(interp, tkImage)
    Tcl_Interp *interp;
    Tk_Image tkImage;
{
    if (tkImage != NULL) {
	if (Tcl_VarEval(interp, "image delete ", Blt_NameOfImage(tkImage), 
			(char *)NULL) != TCL_OK) {
	    return TCL_ERROR;
	}
	Tk_FreeImage(tkImage);
    }
    return TCL_OK;
}

char *
Blt_NameOfImage(tkImage)
    Tk_Image tkImage;
{
    Tk_ImageMaster master;

    master = Tk_ImageGetMaster(tkImage);
    return Tk_NameOfImage(master);
}

/* 
 * --------------------------------------------------------------------------
 *
 * ShearY --
 *
 *	Shears a row horizontally. Antialiasing limited to filtering
 *	two adjacent pixels.  So the shear angle must be between +-45
 *	degrees.
 *	
 * Results:   
 *	None.
 *
 * Side Effects:
 *	The sheared image is drawn into the destination color image.
 *
 * --------------------------------------------------------------------------
 */
static void 
ShearY(src, dest, y, offset, frac, bgColor)
    Blt_Colorimage src, dest;
    int y;			/* Designates the row to be sheared */
    int offset;			/* Difference between  of  */
    double frac;
    Pix32 bgColor;
{
    Pix32 *srcPtr, *destPtr;
    Pix32 *srcRowPtr, *destRowPtr;
    register int x, dx;
    int destWidth;
    int srcWidth;
    double red, blue, green, alpha;
    double leftRed, leftGreen, leftBlue, leftAlpha;
    double oldLeftRed, oldLeftGreen, oldLeftBlue, oldLeftAlpha;

    srcWidth = Blt_ColorimageWidth(src);
    destWidth = Blt_ColorimageWidth(dest);

    destRowPtr = Blt_ColorimageData(dest) + (y * destWidth);
    srcRowPtr = Blt_ColorimageData(src) + (y * srcWidth);

    destPtr = destRowPtr;
    for (x = 0; x < offset; x++) { 
        *destPtr++ = bgColor;
    }
    destPtr = destRowPtr + offset;
    srcPtr = srcRowPtr;
    dx = offset;

    oldLeftRed = (double)bgColor.Red;
    oldLeftGreen = (double)bgColor.Green;
    oldLeftBlue = (double)bgColor.Blue;
    oldLeftAlpha = (double)bgColor.Alpha;

    for (x = 0; x < srcWidth; x++, dx++) { /* Loop through row pixels */
	leftRed = (double)srcPtr->Red * frac;
	leftGreen = (double)srcPtr->Green * frac;
	leftBlue = (double)srcPtr->Blue * frac;
	leftAlpha = (double)srcPtr->Alpha * frac;
        if ((dx >= 0) && (dx < destWidth)) {
	    red = (double)srcPtr->Red - (leftRed - oldLeftRed);
	    green = (double)srcPtr->Green - (leftGreen - oldLeftGreen);
	    blue = (double)srcPtr->Blue - (leftBlue - oldLeftBlue);
	    alpha = (double)srcPtr->Alpha - (leftAlpha - oldLeftAlpha);
	    destPtr->Red = (unsigned char)CLAMP(red);
	    destPtr->Green = (unsigned char)CLAMP(green);
	    destPtr->Blue = (unsigned char)CLAMP(blue);
	    destPtr->Alpha = (unsigned char)CLAMP(alpha);
        }
	oldLeftRed = leftRed;
	oldLeftGreen = leftGreen;
	oldLeftBlue = leftBlue;
	oldLeftAlpha = leftAlpha;
	srcPtr++, destPtr++;
    }
    x = srcWidth + offset;  
    destPtr = Blt_ColorimageData(dest) + (y * destWidth) + x;
    if (x < destWidth) {
	leftRed = (double)bgColor.Red;
	leftGreen = (double)bgColor.Green;
	leftBlue = (double)bgColor.Blue;
	leftAlpha = (double)bgColor.Alpha;

	red = leftRed - (leftRed * frac - oldLeftRed);
	green = leftGreen - (leftGreen * frac - oldLeftGreen);
	blue = leftBlue - (leftBlue * frac - oldLeftBlue);
	alpha = leftAlpha - (leftAlpha * frac - oldLeftAlpha);
	destPtr->Red = (unsigned char)CLAMP(red);
	destPtr->Green = (unsigned char)CLAMP(green);
	destPtr->Blue = (unsigned char)CLAMP(blue);
	destPtr->Alpha = (unsigned char)CLAMP(alpha);
	destPtr++;
    }
    for (x++; x < destWidth; x++) {
        *destPtr++ = bgColor;
    }
}  

/* 
 * --------------------------------------------------------------------------
 *
 * ShearX --
 *
 *	Shears a column. Antialiasing is limited to filtering two
 *	adjacent pixels.  So the shear angle must be between +-45
 *	degrees.
 *	
 * Results:   
 *	None.
 *
 * Side Effects:
 *	The sheared image is drawn into the destination color image.
 *
 * -------------------------------------------------------------------------- 
 */
static void
ShearX(src, dest, x, offset, frac, bgColor)
    Blt_Colorimage src, dest;
    int x;			/* Column in source image to be sheared. */
    int offset;			/* Offset of */
    double frac;			/* Fraction of subpixel. */
    Pix32 bgColor;
{
    Pix32 *srcPtr, *destPtr;
    register int y, dy;
    int destWidth, destHeight;
    int srcWidth, srcHeight;
    double red, blue, green, alpha;
    double leftRed, leftGreen, leftBlue, leftAlpha;
    double oldLeftRed, oldLeftGreen, oldLeftBlue, oldLeftAlpha;
    
    srcWidth = Blt_ColorimageWidth(src);
    srcHeight = Blt_ColorimageHeight(src);
    destWidth = Blt_ColorimageWidth(dest);
    destHeight = Blt_ColorimageHeight(dest);

    destPtr = Blt_ColorimageData(dest) + x;
    for (y = 0; y < offset; y++) {
	destPtr = Blt_ColorimagePixel(dest, x, y);
        *destPtr = bgColor;
	destPtr += destWidth;
    }

    oldLeftRed = (double)bgColor.Red;
    oldLeftGreen = (double)bgColor.Green;
    oldLeftBlue = (double)bgColor.Blue;
    oldLeftAlpha = (double)bgColor.Alpha;
    destPtr = Blt_ColorimageData(dest) + x + offset;
    srcPtr = Blt_ColorimageData(src) + x;
    dy = offset;
    for (y = 0; y < srcHeight; y++, dy++) {
	srcPtr = Blt_ColorimagePixel(src, x, y);
	leftRed = (double)srcPtr->Red * frac;
	leftGreen = (double)srcPtr->Green * frac;
	leftBlue = (double)srcPtr->Blue * frac;
	leftAlpha = (double)srcPtr->Alpha * frac;
        if ((dy >= 0) && (dy < destHeight)) {
	    destPtr = Blt_ColorimagePixel(dest, x, dy);
	    red = (double)srcPtr->Red - (leftRed - oldLeftRed);
	    green = (double)srcPtr->Green - (leftGreen - oldLeftGreen);
	    blue = (double)srcPtr->Blue - (leftBlue - oldLeftBlue);
	    alpha = (double)srcPtr->Alpha - (leftAlpha - oldLeftAlpha);
	    destPtr->Red = (unsigned char)CLAMP(red);
	    destPtr->Green = (unsigned char)CLAMP(green);
	    destPtr->Blue = (unsigned char)CLAMP(blue);
	    destPtr->Alpha = (unsigned char)CLAMP(alpha);
        }
	oldLeftRed = leftRed;
	oldLeftGreen = leftGreen;
	oldLeftBlue = leftBlue;
	oldLeftAlpha = leftAlpha;
	srcPtr += srcWidth; 
	destPtr += destWidth;
    }
    y = srcHeight + offset;  
    destPtr = Blt_ColorimageData(dest) + (y * destWidth) + x + offset;
    if (y < destHeight) {
	leftRed = (double)bgColor.Red;
	leftGreen = (double)bgColor.Green;
	leftBlue = (double)bgColor.Blue;
	leftAlpha = (double)bgColor.Alpha;

	destPtr = Blt_ColorimagePixel(dest, x, y);
	red = leftRed - (leftRed * frac - oldLeftRed);
	green = leftGreen - (leftGreen * frac - oldLeftGreen);
	blue = leftBlue - (leftBlue  * frac - oldLeftBlue);
	alpha = leftAlpha - (leftAlpha  * frac - oldLeftAlpha);
	destPtr->Red = (unsigned char)CLAMP(red);
	destPtr->Green = (unsigned char)CLAMP(green);
	destPtr->Blue = (unsigned char)CLAMP(blue);
	destPtr->Alpha = (unsigned char)CLAMP(alpha);
	destPtr += destWidth;
    }
    for (y++; y < destHeight; y++) {
	destPtr = Blt_ColorimagePixel(dest, x, y);
	*destPtr = bgColor;
	destPtr += destWidth;
    }
}  

/* 
 * ---------------------------------------------------------------------------
 *
 * Rotate45 --
 *
 *	Rotates an image by a given angle.  The angle must be in the
 *	range -45.0 to 45.0 inclusive.  Anti-aliasing filtering is
 *	performed on two adjacent pixels, so the angle can't be so
 *	great as to force a sheared pixel to occupy 3 destination
 *	pixels.  Performs a three shear rotation described below.
 *
 *	Reference: Alan W. Paeth, "A Fast Algorithm for General Raster
 *		Rotation", Graphics Gems, pp 179-195.  
 *
 *
 * Results:  
 *	Returns a newly allocated rotated image.
 *
 * --------------------------------------------------------------------------- 
 */
static Blt_Colorimage
Rotate45(src, theta, bgColor)
    Blt_Colorimage src;
    double theta;
    Pix32 bgColor;
{
    int tmpWidth, tmpHeight;
    int srcWidth, srcHeight;
    double sinTheta, cosTheta, tanTheta;
    double  skewf;
    int skewi;
    Blt_Colorimage tmp1, tmp2, dest;
    register int x, y;

    sinTheta = sin(theta);
    cosTheta = cos(theta);
    tanTheta = tan(theta * 0.5);
    
    srcWidth = Blt_ColorimageWidth(src);
    srcHeight = Blt_ColorimageHeight(src);

    tmpWidth = srcWidth + (int)(srcHeight * FABS(tanTheta));
    tmpHeight = srcHeight;

    /* 1st shear */

    tmp1 = Blt_CreateColorimage(tmpWidth, tmpHeight);
    assert(tmp1);

    if (tanTheta >= 0.0) {	/* Positive angle */
	for (y = 0; y < tmpHeight; y++) {  
	    skewf = (y + 0.5) * tanTheta;
	    skewi = (int)floor(skewf);
	    ShearY(src, tmp1, y, skewi, skewf - skewi, bgColor);
	}
    } else {			/* Negative angle */
	for (y = 0; y < tmpHeight; y++) {  
	    skewf = ((y - srcHeight) + 0.5) * tanTheta;
	    skewi = (int)floor(skewf);
	    ShearY(src, tmp1, y, skewi, skewf - skewi, bgColor);
	}
    }
    tmpHeight = (int)(srcWidth * FABS(sinTheta) + srcHeight * cosTheta) + 1;

    tmp2 = Blt_CreateColorimage(tmpWidth, tmpHeight);
    assert(tmp2);

    /* 2nd shear */

    if (sinTheta > 0.0) {	/* Positive angle */
        skewf = (srcWidth - 1) * sinTheta;
    } else {			/* Negative angle */
        skewf = (srcWidth - tmpWidth) * -sinTheta;
    }
    for (x = 0; x < tmpWidth; x++) {
	skewi = (int)floor(skewf);
        ShearX(tmp1, tmp2, x, skewi, skewf - skewi, bgColor);
	skewf -= sinTheta;
    }

    Blt_FreeColorimage(tmp1);

    /* 3rd shear */

    tmpWidth = (int)(srcHeight * FABS(sinTheta) + srcWidth * cosTheta) + 1;

    dest = Blt_CreateColorimage(tmpWidth, tmpHeight);
    assert(dest);

    if (sinTheta >= 0.0) {	/* Positive angle */
        skewf = (srcWidth - 1) * sinTheta * -tanTheta;
    } else {			/* Negative angle */
        skewf = tanTheta * ((srcWidth - 1) * -sinTheta - (tmpHeight - 1));
    }
    for (y = 0; y < tmpHeight; y++) {
	skewi = (int)floor(skewf);
        ShearY(tmp2, dest, y, skewi, skewf - skewi, bgColor);
	skewf += tanTheta;
    }
    Blt_FreeColorimage(tmp2);
    return dest;      
}

/* 
 * ---------------------------------------------------------------------------
 *
 * CopyColorimage --
 *
 *	Creates a copy of the given color image.  
 *
 * Results:  
 *	Returns the new copy.
 *
 * --------------------------------------------------------------------------- 
 */
static Blt_Colorimage
CopyColorimage(src)
    Blt_Colorimage src;
{
    unsigned int width, height;
    Pix32 *srcPtr, *destPtr;
    Blt_Colorimage dest;

    width = Blt_ColorimageWidth(src);
    height = Blt_ColorimageHeight(src);
    dest = Blt_CreateColorimage(width, height);
    srcPtr = Blt_ColorimageData(src);
    destPtr = Blt_ColorimageData(dest);
    memcpy(destPtr, srcPtr, width * height * sizeof(Pix32));
    return dest;
}

/* 
 * ---------------------------------------------------------------------------
 *
 * Rotate90 --
 *
 *	Rotates the given color image by 90 degrees.  This is part
 *	of the special case right-angle rotations that do not create
 *	subpixel aliasing.
 *
 * Results:  
 *	Returns a newly allocated, rotated color image.
 *
 * --------------------------------------------------------------------------- 
 */
static Blt_Colorimage
Rotate90(src)
    Blt_Colorimage src;
{
    int width, height, offset;
    Pix32 *srcPtr, *destPtr;
    Blt_Colorimage dest;
    register int x, y;

    height = Blt_ColorimageWidth(src);
    width = Blt_ColorimageHeight(src);

    dest = Blt_CreateColorimage(width, height);
    offset = (height - 1) * width;

    srcPtr = Blt_ColorimageData(src);
    offset = (height - 1) * width;
    for (x = 0; x < width; x++) {
	destPtr = Blt_ColorimageData(dest) + offset + x;
	for (y = 0; y < height; y++) {
	    *destPtr = *srcPtr++;
	    destPtr -= width;
	}
    }
    return dest;
}

/* 
 * ---------------------------------------------------------------------------
 *
 * Rotate180 --
 *
 *	Rotates the given color image by 180 degrees.  This is part
 *	of the special case right-angle rotations that do not create
 *	subpixel aliasing.
 *
 * Results:  
 *	Returns a newly allocated, rotated color image.
 *
 * --------------------------------------------------------------------------- 
 */
static Blt_Colorimage
Rotate180(src)
    Blt_Colorimage src;
{
    int width, height, offset;
    Pix32 *srcPtr, *destPtr;
    Blt_Colorimage dest;
    register int x, y;

    width = Blt_ColorimageWidth(src);
    height = Blt_ColorimageHeight(src);
    dest = Blt_CreateColorimage(width, height);

    srcPtr = Blt_ColorimageData(src);
    offset = (height - 1) * width;
    for (y = 0; y < height; y++) {
	destPtr = Blt_ColorimageData(dest) + offset + width - 1;
	for (x = 0; x < width; x++) {
	    *destPtr-- = *srcPtr++;
	}
	offset -= width;
    }
    return dest;
}

/* 
 * ---------------------------------------------------------------------------
 *
 * Rotate270 --
 *
 *	Rotates the given color image by 270 degrees.  This is part
 *	of the special case right-angle rotations that do not create
 *	subpixel aliasing.
 *
 * Results:  
 *	Returns a newly allocated, rotated color image.
 *
 * --------------------------------------------------------------------------- 
 */
static Blt_Colorimage
Rotate270(src)
    Blt_Colorimage src;
{
    int width, height;
    Pix32 *srcPtr, *destPtr;
    Blt_Colorimage dest;
    register int x, y;

    height = Blt_ColorimageWidth(src);
    width = Blt_ColorimageHeight(src);
    dest = Blt_CreateColorimage(width, height);

    srcPtr = Blt_ColorimageData(src);
    for (x = width - 1; x >= 0; x--) {
	destPtr = Blt_ColorimageData(dest) + x;
	for (y = 0; y < height; y++) {
	    *destPtr = *srcPtr++;
	    destPtr += width;
	}
    }
    return dest;
}

/*
 *----------------------------------------------------------------------
 *
 * Blt_RotateColorimage --
 *
 *	Rotates a color image.
 *
 * Results:
 *	The color image is converted to greyscale.
 *
 *----------------------------------------------------------------------
 */
Blt_Colorimage
Blt_RotateColorimage(src, angle)
    Blt_Colorimage src;
    double angle;
{
    Blt_Colorimage dest, tmp;
    int quadrant;
    
    tmp = src;			/* Suppress compiler warning. */

    /* Make the angle positive between 0 and 360 degrees. */ 
    angle = FMOD(angle, 360.0);
    if (angle < 0.0) {
	angle += 360.0;
    }
    quadrant = ROTATE_0;
    if ((angle > 45.0) && (angle <= 135.0)) {
        quadrant = ROTATE_90;
        angle -= 90.0;
    } else if ((angle > 135.0) && (angle <= 225.0)) { 
        quadrant = ROTATE_180;
        angle -= 180.0;
    } else if ((angle > 225.0) && (angle <= 315.0)) { 
        quadrant = ROTATE_270;
        angle -= 270.0;
    } else if (angle > 315.0) {
	angle -= 360.0;
    }
    /* 
     * If necessary, create a temporary image that's been rotated
     * by a right-angle.  We'll then rotate this color image between
     * -45 to 45 degrees to arrive at its final angle.  
     */
    switch (quadrant) {
    case ROTATE_270:		/* 270 degrees */
	tmp = Rotate270(src);
	break;

    case ROTATE_90:		/* 90 degrees */
	tmp = Rotate90(src);
	break;

    case ROTATE_180:		/* 180 degrees */
	tmp = Rotate180(src);
	break;

    case ROTATE_0:		/* 0 degrees */
	if (angle == 0.0) {
	    tmp = CopyColorimage(src); /* Make a copy of the source. */
	} 
	break;
    }

    dest = tmp;
    assert((angle >= -45.0) && (angle <= 45.0));
    if (angle != 0.0) {
	double theta;
	Pix32 *srcPtr;
	Pix32 bgColor;

	/* FIXME: pick up background blending color from somewhere */
	srcPtr = Blt_ColorimageData(src);
	bgColor = *srcPtr;
	bgColor.Red = bgColor.Green = bgColor.Blue = 0xFF;
	bgColor.Alpha = 0x00;	/* Transparent background */
	theta = (angle / 180.0) * M_PI;
	dest = Rotate45(tmp, theta, bgColor);
	if (tmp != src) {
	    Blt_FreeColorimage(tmp);
	}
    } 
    return dest;
}
