########################################################################
#
# balloonHelp -- attach helpful popup balloon to any widget
#
# 
# USAGE 
# balloonHelp client text
#   client -- an existing widget
#   text   -- the helpful text
#
# GLOBAL VARIABLES
# balloonHelp -- messages only pop up, if this is equal to 1
# _ballonHelp -- none of your business
#
# WIDGETS
# .balloonHelp -- the toplevel widget displaying the help text
#
# EXAMPLE
# button .b -text save
# balloonHelp .b "saves your data in a file you specify"
#
#
# DESCRIPTION
# BallonHelp extends the <Enter>-bindings of the client widget such
# that an enter-event initiates the popup of a message window with the
# given text. However, the popup does not show up immediately, but
# after a short delay, given the the mouse was not moved in the
# meantime. The idea is, that if the user is roaming around the widget
# with the mouse pointer, (s)he know what to do and needs no
# help. However, if (s)he pause in hesitation, the help-message pops
# up.
# The <Leave>-event of the client widget is extendet to pop down the
# message window.
#
# HINTS
# o Don't call balloonHelp before you fixed all <Enter>- or
#   <Leave>-bindings for a client widget, or call balloonHelp again
#   after changing a binding.
# o to effectively turn off balloonHelp, set the global variable
#   balloonHelp to anything but 0
#
#
# (C) 1997 Harald Kirsch (kir@iitb.fhg.de)
# $Revision: 1.3 $, $Date: 1997/05/24 14:04:41 $
#
########################################################################

## default is to turn on balloonHelp
set balloonHelp 1

########################################################################
#
# balloonHelp -- this is the API
#
#   w -- the client widget 
# msg -- the message text to bind to the client
#
proc balloonHelp {w msg} {
  bind $w <Enter> "+balloonHelp.init %W {$msg} %t"
  bind $w <Leave> "+balloonHelp.down %W"
}
########################################################################
#
# balloonHelp.init -- local function, bound to <Enter>-events of
#                     clients
#
proc balloonHelp.init {client msg time} {
  #puts [bind $client]
  global _balloonHelp balloonHelp

  if { !$balloonHelp } return

  set x [winfo pointerx $client]
  set y [winfo pointery $client]
  set _balloonHelp(itime) $time
  #puts "initiating chain $time"
  after 400 balloonHelp.up $client "{$msg} $time $x.$y 3"
}
########################################################################
#
# balloonHelp.up -- local function, called after delay to pop up the
#                   message
#
proc balloonHelp.up {client msg itime pos try} {
  global _balloonHelp balloonHelp

  set x [winfo pointerx $client]
  set y [winfo pointery $client]

  ## If balloonHelp was switched off in the meantime, if .init was
  ## called again or if we tried often enough, nothing is to be done
  if { !$balloonHelp || "$itime"!="$_balloonHelp(itime)" || $try<=0 } {
    #puts "stopping chain $itime"
    return
  }
  incr try -1

  ## If the mouse was moved, nothing is to be done, but we try again
  ## later.
  if { "$pos"!="$x.$y" } {
    #puts "extending chain $_balloonHelp(itime)"
    after 1000 \
	balloonHelp.up $client "{$msg}" \
	$_balloonHelp(itime) $x.$y $try
    return
  }

  ## fire it up
  set x [winfo rootx $client]
  set y [winfo rooty $client]
  set x [expr $x + [winfo width $client]/2]
  set y [expr $y + [winfo height $client]+2]
  #set x [expr $x+10]
  #set y [expr $y-10]

  catch {destroy .balloonHelp}
  toplevel .balloonHelp -borderwidth 2 -background black
  wm overrideredirect .balloonHelp 1
  wm geometry .balloonHelp +$x+$y
  
  message .balloonHelp.m \
      -aspect 1000 \
      -relief flat \
      -background gold \
      -foreground  black \
      -text $msg
  #E03010
  pack .balloonHelp.m
}
########################################################################
proc balloonHelp.down {w} {
  global _balloonHelp
  set _balloonHelp(itime) x
  catch {destroy .balloonHelp}
}
########################################################################
if 0 {

  frame .f
  checkbutton .f.cb  \
      -text "balloon help on/off" \
      -variable balloonHelp

  button .f.b -text hallo \
      -command "puts hallo"
  canvas .f.c -background \#334488

  menubutton .f.m -menu .f.m.menu -text File
  menu .f.m.menu
  .f.m.menu add command -label {Open ...}
  .f.m.menu add command -label {Save}

  bind .f.c <Enter> {%W configure -cursor gumby}
  bind .f.c <Leave> {%W configure -cursor {}}

  pack .f.cb .f.b .f.m .f.c
  pack .f

  balloonHelp .f.c "Dies ist der Canvas"
  balloonHelp .f.b "Ein total witziger Hallo-Knopf.\nDer tut nichts."
  balloonHelp .f.cb \
      "Wenn Sie diesen Knopf umschalten, erscheinen die netten\
kleinen\
Hilfestellungen wie diese hier nicht mehr, also Vorsicht!"
  balloonHelp .f.m.menu "You dummy!"
  balloonHelp .f.m "Alle mglichen Operationen auf Dateien"
}
