########################################################################
# Tdf related functions:
#
# tdf::ReadHeader 
# tdf::ReadData
# tdf::Read
#
# This makes heavy use of the `binary'-command of tcl8.0, so it will not
# work with earlier versions.
#
# $Revision: 1.4 $, $Date: 1998/11/01 18:26:08 $
########################################################################

namespace eval tdf {
  namespace export ReadHeader ReadData

  set trans(C,0)  a
  set trans(C,1)  a
  set trans(S,0)  s
  set trans(S,1)  S
  set trans(US,0) s
  set trans(US,1) S
  set trans(I,0)  i
  set trans(I,1)  I
  set trans(UI,0) i
  set trans(UI,1) I
  set trans(F,0)  f
  set trans(F,1)  f
  set trans(D,0)  d
  set trans(D,1)  d
  
  set sizeof(C)  1
  set sizeof(S)  2
  set sizeof(US) 2
  set sizeof(I)  4
  set sizeof(UI) 4
  set sizeof(F)  4
  set sizeof(D)  8
}
######################################################################
#
# tdf::ReadHeader -- read tdf-header from an open file
#
# PARAM:
# in -- handle of an open file
# _header -- name of a variable. On return, it will be an array
#            with one element for every header entry.
#
# RETURN: 
# There is no return value. On error, a tcl-error is generated.
#
proc tdf::ReadHeader {in _header} {
  upvar $_header header
  
  set i 0
  while { ![eof $in] && "\025"!=[set line [read $in 1]] } {
    incr i
    if {$line=="\n"} continue
    gets $in l
    regsub "\r$" $l {} l	;#string \r from end of line
    append line $l
    
    ## chop line into keyword and value
    if ![regexp { *([^: ]+) *:} $line dummy varname] {
      return -code error -errorinfo {} \
	  "missing `:' in header line $i"
    }
    regexp {: *(.*)} $line dummy varvalue
    set varvalue [string trim $varvalue]
    set header($varname) $varvalue
  }
  if {"\025"!=$line} {
    error "no binary marker found in file `$fName'"
  }
}

######################################################################
#
# splitFormat - chop off type and count from the head of a format
# string (private function)
#
# 
#
proc tdf::splitFormat {format _count _type _rest} {
  upvar $_count count $_type type $_rest rest
  
  if {![regexp { *([0-9]+)(C|US|S|UI|I|F|D|\()(.*)} \
	    $format dummy count type rest]} {
    error "syntax error in format at suffix `$format'"
  }
  set rest [string trim $rest]
}
######################################################################
#
# _read - read data from open file according to format (private
# function) 
#
proc tdf::_read {in format isBigEndian _rest} {
  upvar $_rest rest
  variable sizeof
  variable trans
  global tcl_platform

  #puts "_read: `$format'"
  set res {}

  ##
  ## read on, while there is something left in the format string and
  ## it does not start with ')'
  ##
  while {"$format"!="" && ![string match "\)*" $format]} {
    splitFormat $format count type rest
    #puts "$format -> `$count'  `$type'  `$rest'"
    
    if {"$type"=="("} {
      ##
      ## Handle subformat introduced by "("
      ##
      set elem {}
      set format $rest
      for {set i 0} {$i<$count} {incr i} {
	set elem [concat $elem [_read $in $format $isBigEndian rest]]
      }
      if {![string match "\)*" $rest]} {
	error "missing close brace after `$format'"
      }
      set rest [string range $rest 1 end]
      #puts "--$format--"
    } else {

      ##
      ## Read in pure data
      ##
      set size [expr $count*$sizeof($type)]
      set buf [read $in $size]
      if {[string length $buf]!=$size} {
	error "short file at format suffix `$format'"
      }

      ## swap bytes for F and D, if necessary
      if {("$type"=="F" || "$type"=="D")
	  && ($isBigEndian && "$tcl_platform(byteOrder)"!="bigEndian")} {
	byteswap buf $sizeof($type)
      }

      #puts "scanning: $trans($type,$isBigEndian)$count"
      binary scan $buf $trans($type,$isBigEndian)$count elem
      #puts $elem

      ## correct unsigned values
      switch -exact -- $type {
	US {
	  set tmp $elem
	  set elem {}
	  foreach x $tmp {
	    lappend elem [expr ($x+0x100)%0x100]
	  }
	}
	UI {
	  set tmp $elem
	  set elem {}
	  set phi [expr pow(2,32)]
	  foreach x $tmp {
	    lappend elem [expr $x>=0 ? $x : $phi+$x]
	  }
	}
      }
    }

    lappend res [list $type $elem]
    set format $rest
  }
  return $res
}
######################################################################
#
# tdf::ReadData - read data from open file
#
# PARAMETERS:
# in -- file handle
# _header -- name of an array variable containing the header as it was
#     read in with tdf::ReadHeader
#
# RETURN:
# A (recursive) list containing the data as it was read from the
# file. Every list element contains a type-tag and the data. For
# example a format element 2F is translated into a pair like
#   {F {1.0 1.0}
# Every type tag of the format string results in a list element. In
# particular a subformat like 2(2F 3I) results in one list element on
# the top level. This list element then contains again a list with the
# elements of the subformat. The result for the example above would
# contain 4 elements like 
# {F {1.0 1.0}} {I {1 1 1}} {F {2.0 2.0}} {I {2 2 2}}
#
proc tdf::ReadData {in _header} {
  upvar $_header header
  global tcl_platform

  if {[info exist header(BigEndian)]} {
    set isBigEndian 1
  } elseif {[info exist header(LittleEndian)]} {
    set isBigEndian 0
  } else {
    set isBigEndian [expr \"$tcl_platform(byteOrder)\" == \"bigEndian\" ]
  }

  ## This is supposed to be **BINARY** data, so we don't need no
  ## steenking translation of any MicroShit kind
  fconfigure $in -translation binary

  set res [_read $in $header(Format) $isBigEndian rest]
  if {"$rest"!=""} {
    error "uninterpreted rest `$rest' in format string"
  }
  return $res;
}
########################################################################
proc tdf::Read {fname _header} {
  upvar $_header header

  set in [open $fname r]
  tdf::ReadHeader $in header
  set res [tdf::ReadData $in header]
  close $in
  return $res
}
########################################################################

if 0 {
  proc show {res ind} {
    foreach x $res {
      if {"[lindex $x 0]"=="("} {
	puts "$ind\("
	show [lindex $x 1] "$ind  "
      } else {
	puts "$ind$x"
      }
    }
  }
  
  set res [tdf::Read [lindex $argv 0] header]
  parray header
  show $res {}
  exit

  set in [open [lindex $argv 0] r]
  tdf::ReadHeader $in header
  parray header
  
  set res [tdf::ReadData $in header]
  #puts $res
  show $res {}
}
