########################################################################
#
# $Revision: 2.11 $ $Date: 2000/01/13 09:26:28 $
#
########################################################################
##
## We maintain information about the files associated with a
## nep-window. All access goes through the file's handle which is used
## as an index into an array.

namespace eval tdfio {
  variable I;				# this contains everything

  variable defaultHeader;
  set defaultHeader(Increment) 1.0
  set defaultHeader(Origin) 0.0
  set defaultHeader(ConversionFactor) 1.0
  set defaultHeader(ConversionOffset) 0.0
  set defaultHeader(Invert) 0

  variable typetrans
  set typetrans(S) i2
  set typetrans(US) u2
  set typetrans(I) i4
  set typetrans(UI) u4
  set typetrans(F) r4
  set typetrans(D) r8
}


########################################################################
#
# Check, if this computer is a big-endian machine. Since Tcl8.0 this
# is trivial. For other versions of Tcl use an older version of nep!
# 
proc ::tdfio::isBigEndian {} {
  global tcl_platform
  if [info exist tcl_platform(byteOrder)] {
    return [expr \"$tcl_platform(byteOrder)\" != \"littleEndian\" ]
  }
}
########################################################################
proc ::tdfio::getCount {fd} {
  variable I
  return $I($fd,count)
}
proc ::tdfio::getName {fd} {
  variable I
  return $I($fd,name)
}
proc ::tdfio::getRaw {fd} {
  variable I
  return $I($fd,raw)
}
########################################################################
proc ::tdfio::get {fd key {defalt {}}} {
  variable I

  if {[llength [info level 0]]==4} {
    ## default value was actually given by caller
    if {![info exist I($fd,h,$key)]} {
      return $defalt
    }
  }
  return $I($fd,h,$key)
}
########################################################################
proc ::tdfio::hset {fd key value} {
  variable I
  set I($fd,h,$key) $value
}
########################################################################
## read tdf header info from file f (which has name fname, used only
## for error messages)
##
proc ::tdfio::readHeader {f fName} {
  variable I
  variable defaultHeader

  ## Read all data up to the binary marker into h without split.
  set h {}
  set hlen 0
  set hok 0
  while { ![eof $f] } {
    set buf [::read $f 512]
    set pos [string first "\025" $buf]
    if {$pos==-1} {
      append h $buf
      incr hlen 512
    } else {
      incr pos -1;			# don't want the \025
      append h [string range $buf 0 $pos]
      incr hlen $pos
      incr hlen
      set hok 1
      break
    }
  }

  if {!$hok} {
    error "no binary marker found in file `$fName'"
  }

  ## Now split into lines and extract key and value. An old definition
  ## says that the key starts at the first non-blank character after the 
  ## colon (if any)
  foreach line [split $h "\n"] {
    set key {}
    set value {}
    regexp { *([^ :]+) *:? *(.*)} $line {} key value
    set I($f,h,$key) $value
  }

  ## fill in some defaults which we always need
  foreach x [array names defaultHeader] {
    if {![info exist I($f,h,$x)]} {
      set I($f,h,$x) $defaultHeader($x)
    }
  }
  if {![info exist I($f,h,Comment)]} {
    set I($f,h,Comment) [file tail $fName]
  }

  ## decide if we have to swap input words
  if {([info exist I($f,h,BigEndian)] && ![isBigEndian])
      || ([info exist I($f,h,LittleEndian)] && [isBigEndian])} {
    set I($f,swap) 1
  } else {
    set I($f,swap) 0
  }
}
##########################################################################
proc ::tdfio::setHlen {fd} {
  variable I

  seek $fd 0 start
  fconfigure $fd -translation binary
  set hlen 0
  while { ![eof $fd]} {
    set buf [::read $fd 512]
    set pos [string first "\025" $buf]
    if {$pos==-1} {
      append h $buf
      incr hlen 512
    } else {
      incr hlen $pos
      incr hlen
      break
    }
  }
  set I($fd,hlen) $hlen
}
##########################################################################
## We allow quite a few formats. In particular, we require only the
## front of the format string to match something useful. The rest is
## silently ignored.
##
## The following formats are allowed:
##
##   *S *US *I *UI *F *D
##   *(*S) *(*US) *(*I) *(*UI) *(*F) *(*D)
##   nSnS nUSnUS nInI nUInUI nFnF
##
## If the format describes a matrix (like in the second line), the
## `Blocksize'-entry of tdfheader is set to the length of a line of
## the matrix, i.e. to the count within the parentheses.
##
## A format like nTnT, where T is any type, is recognized for a file
## which contains x- and y-coordinates. 
##
## For all other formats, the x-coordinates may come from a file
## mentioned in tdfheader(Xaxisfile).
##
proc ::tdfio::parseFormat {fd} {
  variable I
  variable typetrans

  set anytag (S|US|I|UI|F|D)
  set int {[0-9]+}
  set ws "\[ \t\n]*"

  # parray I
  if {![info exist I($fd,h,Format)]} {
    error "no `Format' found in header of $I($fd,name)"
  }
  set format [string trim $I($fd,h,Format)]

  ## Try to match `nTnT' for any T listed above
  if {[regexp "^($int)$anytag${ws}($int)$anytag" \
	   $format dummy c1 t1 c2 t2]
      && $c1==$c2 && "$t1"=="$t2"} {
    return -code error "net yet implemented"
    set I($fd,haveX) 1
    set I($fd,count) [expr $c1+$c2]
    set I($fd,type) $typetrans($t1)

  ## Try to match `nT' for any T listed above
  } elseif {[regexp "^($int)$anytag" $format dummy count type]} {
    set I($fd,haveX) 0
    set I($fd,count) $count
    set I($fd,type) $typetrans($type)

    ## Try to match `m(nT)' for any T listed above
  } elseif {[regexp "^($int)\\(${ws}($int)$anytag\\)" \
		  $format dummy c1 c2 type]} {
    set I($fd,haveX) 0
    set I($fd,count) [expr $c1*$c2]
    set I($fd,Blocksize) $c2
    set I($fd,type) $typetrans($type)

  } else {
    close $fd
    foreach name [array names I $fd,*] {
      unset I($name)
    }
    error "wrong format in file $fName"
  }

}
########################################################################
#
# PARAMETERS:
# xaxisok --
# should never be used by clients. It is used internally to signal that
# it is useless to look for an Xaxisfile if we are already loading an
# Xaxisfile. 
proc ::tdfio::open {fname raw {xaxisok 1}} {
  variable I

  set fd [::open $fname r]

  set I($fd,name) $fname
  set I($fd,raw) $raw

  ## These functions set elements of I
  readHeader $fd $fname
  if {[info exist I($fd,h,Ascii)]} {
    return -code error "ascii data no longer supported"
  }
  setHlen $fd
  parseFormat $fd

  ## set the sizeof entry, I($fd,type) is a type of blt like u4
  set I($fd,sizeof) [string range $I($fd,type) 1 end]

  ## the next lines of magic are necessary because we want to
  ## switch to binary translation.
  seek $fd $I($fd,hlen) start
  set I($fd,hlen) [tell $fd]
  fconfigure $fd -translation binary

  ## If there is a header entry Xaxisfile, prepare the xaxis
  if {$xaxisok && [info exist I($fd,h,Xaxisfile)]} {
    set I($fd,xfd) [openXaxisfile $fd $I($fd,h,Xaxisfile)]
  }

  return $fd
}
########################################################################
#
# this is a local function and should not be used by clients.
#
proc ::tdfio::openXaxisfile {yfd xaf} {
  variable I

  switch -exact [file pathtype $I($yfd,name)] {
    relative {
      set xaf [file join [file dir $I($yfd,name)] $xaf]
    }
    default {
      #relax
    }
  }
  return [open $xaf 0 0]
}
########################################################################
proc ::tdfio::xCoord {fd s} {
  variable I
  if {"$s"=="last"} { set s $I($fd,count) ; incr s -1}

  if {[info exist I($fd,xfd)]} {
    vector v
    getSample $I($fd,xfd) $s v
    return $v(0)
  } 

  return [expr {$s*$I($fd,h,Increment)+$I($fd,h,Origin)}]
}
########################################################################
#
# Read sample $i from $fd into vector $v at index 0.
#
proc ::tdfio::getSample {fd i v} {
  variable I
  seek $fd [expr {$i*$I($fd,sizeof)+$I($fd,hlen)}] start
  if {$I($fd,swap)} {
    $v binread $fd 1 -format $I($fd,type) -swap -at 0
  } else {
    $v binread $fd 1 -format $I($fd,type) -at 0
  }
}
########################################################################
#
# This is a binary search to find the largest index into the given
# signal such that the sample value is still smaller than the given
# x. Of course it is assumed that the file is sorted.
# If all samples are greater than x, 0 is returned.

proc ::tdfio::bsearchFile {fd x} {
  variable I
  vector v

  set lo 0
  set hi $I($fd,count)

  while {$lo+1<$hi} {
    set mid [expr {($lo+$hi)/2}]
    
    getSample $fd $mid v
    #puts "$lo,$mid,$hi -> $v(0), looking for $x"

    if {$v(0)<$x} {
      set lo $mid
    } elseif {$x<$v(0)} {
      set hi $mid
    } else {
      return [list $mid $v(0)]
    }
  }

  getSample $fd $lo v
  
  return [list $lo $v(0)]
}
########################################################################
##
## find out which sample lies just right or on the given coordinate
## If no such sample exists, I($fd,count)-1 is returned, otherwise the
## sample number.
proc ::tdfio::rSample {fd x} {
  variable I

  if {[info exist I($fd,xfd)]} {
    foreach {i v} [bsearchFile $I($fd,xfd) $x] break
    if {$i==$I($fd,count)-1} {return $i}
    if {$v>$x} {return $i}
    incr i
    return $i
  }
    

  set o $I($fd,h,Origin)
  if {$x<=$o} {return 0}

  set i [expr {int(ceil(($x-$o)/$I($fd,h,Increment)))}]
  if {$i>=$I($fd,count)} {return [expr {$I($fd,count)-1}]}

  return $i
}
########################################################################
## find out which sample lies just left or on the given coordinate
## If no such sample exists, 0 is returned, otherwise the
## sample number.
proc ::tdfio::lSample {fd x} {
  variable I
  
  if {[info exist I($fd,xfd)]} {
    return [lindex [bsearchFile $I($fd,xfd) $x] 0]
  }

  set o $I($fd,h,Origin)
  if {$x<=$o} {return 0}

  set i [expr {int(floor(($x-$o)/$I($fd,h,Increment)))}]
  if {$i>=$I($fd,count)} {
    set i [expr {$I($fd,count)-1}]
  }

  return $i
}
########################################################################
proc ::tdfio::read {fd xfrom xto xvec yvec} {
  variable I

  ## The current contents are no longer needed
  $xvec set {}
  $yvec set {}
  #puts >>[info level 0]
  set ls [lSample $fd $xfrom]
  set rs [rSample $fd $xto]

  set count [expr {$rs-$ls+1}]

  #puts "ls=$ls rs=$rs ($xfrom,$xto)"

  ## read the y-data
  seek $fd [expr {$ls*$I($fd,sizeof)+$I($fd,hlen)}] start
  if {$I($fd,swap)} {
    $yvec binread $fd $count -format $I($fd,type) -swap -at 0
  } else {
    $yvec binread $fd $count -format $I($fd,type) -at 0
  }

  ## apply scale, offset and inversion, if necessary
  #puts [$yvec range 0 end]
  if {!$I($fd,raw)} {
    $yvec expr \
	"$yvec * $I($fd,h,ConversionFactor) + $I($fd,h,ConversionOffset)"
    if {$I($fd,h,Invert)} {
      $yvec expr "$yvec ^ (-1)"
    }
  }

  ## make the x-axis
  if {[info exist I($fd,xfd)]} {
    set xfd $I($fd,xfd)
    seek $xfd [expr {$ls*$I($xfd,sizeof)+$I($xfd,hlen)}] start
    if {$I($xfd,swap)} {
      $xvec binread $xfd $count -format $I($xfd,type) -swap -at 0
    } else {
      $xvec binread $xfd $count -format $I($xfd,type) -at 0
    }
  } elseif {$I($fd,haveX)} {
    ## not yet implemented
  } else {
    set xmin [expr {$I($fd,h,Origin)+$ls*$I($fd,h,Increment)}]
    set xmax [expr {$I($fd,h,Origin)+$rs*$I($fd,h,Increment)}]
    $xvec seq $xmin $xmax $I($fd,h,Increment)
    #puts "xvec has length [$xvec length] ($xmin, $xmax)"
  }
}
########################################################################
proc ::tdfio::close {fd} {
  variable I

  ::close $fd
  if {[info exist I($fd,xfile)]} { close $I($fd,xfile) }
    
  foreach x [array names I $fd,*] {
    unset I($x)
  }
}
########################################################################
